import numpy as np 

from parakeet.testing_helpers import run_local_tests, expect  


int_array = np.array([0,1])
bool_array = np.array([False,True])
float_array = np.array([0.1, 0.2])
arrays = [int_array, bool_array, float_array]

def unary(fn, data = None): 
  for x in (arrays if data is None else data):
    expect(fn, [x], fn(x))

def binary(fn):
  for x in arrays:
      for y in arrays:
          expect(fn, [x,y], fn(x,y))

def test_sin():
  unary(np.sin)

def test_cos():
  unary(np.cos)
  
def test_tan():
  unary(np.tan)
  
def test_arcsin():
  unary(np.arcsin)
  
def test_arccos():
  unary(np.arccos)
  
def test_arctan():
  unary(np.arctan)
  
def test_arctan2():
  binary(np.arctan2)
  
  
def test_sinh():
  unary(np.sinh)
  
def test_cosh():
  unary(np.cosh)
  
def test_tanh():
  unary(np.tanh)
  
def test_arcsinh():
  unary(np.arcsinh)
  
def test_arccosh():
  unary(np.arccosh, data = np.array([2.0, 3.0]))
  
def test_arctanh():
  unary(np.arctanh, data = np.array([0.2, 0.3]))
  
def test_deg2rad():
  unary(np.deg2rad)
  
def test_rad2deg():
  unary(np.rad2deg)
  
#def test_hypot():
#  binary(np.hypot)

if __name__ == '__main__':
  run_local_tests() 



from ..frontend import jit 

import numpy as np 



@jit
def conjugate(x):
  """
  For now we don't have complex numbers so this is just the identity function
  """
  return x


@jit
def real(x):
  """
  For now we don't have complex types, so real is just the identity function
  """
  return x  

def _scalar_sign(x):
  if x > 0:
    return 1
  elif x < 0:
    return -1 
  else:
    return 0
  
@jit
def sign(x):
  return map(_scalar_sign, x)

@jit 
def reciprocal(x):
  return 1 / x

@jit 
def rad2deg(rad):
  return rad * 180 / 3.141592653589793

@jit
def deg2rad(deg):
  return deg * 3.141592653589793 / 180 

@jit 
def hypot(x,y):
  return np.sqrt(x**2 + y**2)

@jit 
def square(x):
  return x * x 

def _logaddexp_scalar(x, y):
  """
  Copied from BioPython (http://biopython.org/)
  """
  if x < y:
    bigger = x 
    smaller = y 
  else:
    bigger = x 
    smaller = y 
    
  diff = smaller - bigger    
  if diff < -100: 
    return bigger 

  return bigger + np.log1p(np.exp(diff))
  

@jit 
def logaddexp(x, y):
  return map(_logaddexp_scalar, x, y)

@jit   
def log2_1p(x):
  return (1.0 / np.log(2)) * np.log1p(x)


@jit 
def logaddexp2(x, y):
  diff = x - y 
  return np.where(diff > 0,  x + log2_1p(2 ** -diff) , y + log2_1p(2 ** diff))

 
@jit 
def true_divide(x, y):
  """
  Not exactly true divide, since I guess it's sometimes supposed to stay an int
  """
  return (x + 0.0) / (y + 0.0)

@jit 
def floor_divide(x, y):
  return np.floor(x / y)  



from .. import config 

from ..c_backend.prepare_args import prepare_args  
from ..transforms.pipeline import after_indexify, final_loop_optimizations  
from ..value_specialization import specialize


from multicore_compiler import MulticoreCompiler 

_cache = {}
def run(fn, args):
  args = prepare_args(args, fn.input_types)
  
  fn = after_indexify(fn)
  fn = final_loop_optimizations.apply(fn)
  if config.value_specialization:
    fn = specialize(fn, python_values = args)
  key = fn.cache_key 
  if key in _cache:
    return _cache[key](*args)
  else:
    compiled_fn = MulticoreCompiler().compile_entry(fn)
    c_fn = compiled_fn.c_fn 
    _cache[key] = c_fn 
    return c_fn(*args)
  


     
import numpy as np 
def summation(pos, weights, points):
  n_points = len(points)
  n_weights = len(weights)
  sum_array = np.zeros(n_points)
  sum_array3d = np.zeros((n_points,3))
  def compute(i):
    pxi = points[i, 0]
    pyi = points[i, 1]
    pzi = points[i, 2]
    total = 0.0
    for j in xrange(n_weights):
      weight_j = weights[j]
      xj = pos[j,0]
      yj = pos[j,1]
      zj = pos[j,2]
      dx = pxi - pos[j, 0]
      dy = pyi - pos[j, 1]
      dz = pzi - pos[j, 2]
      dr = 1.0/np.sqrt(dx*dx + dy*dy + dz*dz)
      total += weight_j * dr
      sum_array3d[i,0] += weight_j * dx
      sum_array3d[i,1] += weight_j * dy
      sum_array3d[i,2] += weight_j * dz
    return total 
  sum_array = np.array([compute(i) for i in xrange(n_points)])
  return sum_array, sum_array3d

n_points = 200
n_weights = 400
pos = np.random.randn(n_weights, 3)
weights = np.random.randn(n_weights)
points = np.random.randn(n_points, 3)

from compare_perf import compare_perf 

compare_perf(summation, [pos, weights, points])





from .. syntax import (none, zero_i64, 
                       Map, Reduce, Scan, 
                       IndexMap, IndexReduce, 
                       ParFor, OuterMap,
                       Filter, FilterReduce) 

from .. frontend import macro, staged_macro, jit,  translate_function_value 
from lib_helpers import _get_shape 
@jit 
def identity(x):
  return x

@macro 
def parfor(fn, bounds):
  fn = translate_function_value(fn)
  bounds = _get_shape(bounds)
  return ParFor(fn = fn, bounds = bounds)

@staged_macro("axis")
def map(f, *args, **kwds):
  if 'axis' in kwds:
    axis = kwds['axis']
    del kwds['axis']
  else:
    axis = zero_i64
  assert len(kwds) == 0, "map got unexpected keywords %s" % (kwds.keys())
  return Map(fn = f, args = args, axis = axis)
each = map

@staged_macro("axis")
def allpairs(f, x, y, axis = 0):
  return OuterMap(fn = f, args = (x,y), axis = axis)

@staged_macro("axis")
def outer_map(f, *args, **kwds):
  if 'axes' in kwds:
    assert 'axis' not in kwds, "Can't have both 'axis' and 'axes' as keywords"
    axis = kwds['axes']
    del kwds['axes'] 
    
  elif 'axis' in kwds:
    axis = kwds['axis']
    del kwds['axis']
  else:
    axis = zero_i64
  assert len(kwds) == 0, "outer_map got unexpected keywords %s" % (kwds.keys())
  return OuterMap(fn = f, args = args, axis = axis)

@staged_macro("axis")
def reduce(f, *args, **kwds):
  if 'axis' in kwds:
    axis = kwds['axis']
    del kwds['axis']
  else: 
    axis = none
  if 'init' in kwds:
    init = kwds['init']
    del kwds['init']
  else:
    init = none 
  assert len(kwds) == 0, "reduce got unexpected keywords %s" % (kwds.keys())
  
  ident = translate_function_value(identity)
  return Reduce(fn = ident, 
                combine = f, 
                args = args,
                init = init,
                axis = axis)

@staged_macro("axis")
def scan(f, *args, **kwds):
  axis = kwds.get('axis', zero_i64)
  init = kwds.get('init', none)
  
  ident = translate_function_value(identity)
  return Scan(fn = ident,  
                     combine = f,
                     emit = ident, 
                     args = args,
                     init = init,
                     axis = axis)
@macro
def imap(fn, shape):
  return IndexMap(shape = shape, fn = fn)


@macro
def ireduce(fn, combine, shape, init = None):
  return IndexReduce(fn = fn, combine=combine, shape = shape, init = init)

@staged_macro("axis")
def filter(pred, *args, **kwds):
  if 'axis' in kwds:
    axis = kwds['axis']
    del kwds['axis']
  else:
    axis = zero_i64
  assert len(kwds) == 0, "filter got unexpected keywords %s" % (kwds.keys())
  return Filter(fn = pred, args = args, axis = axis)


@staged_macro("axis")
def filter_reduce(f, pred, *args, **kwds):
  if 'axis' in kwds:
    axis = kwds['axis']
    del kwds['axis']
  else: 
    axis = none
  if 'init' in kwds:
    init = kwds['init']
    del kwds['init']
  else:
    init = none 
  assert len(kwds) == 0, "reduce got unexpected keywords %s" % (kwds.keys())
  
  ident = translate_function_value(identity)
  return FilterReduce(fn = ident, 
                      pred = pred, 
                      combine = f, 
                      args = args,
                      init = init,
                      axis = axis)



import numpy as np 
from parakeet import jit 

@jit 
def fdtd(input_grid, steps):
    grid = input_grid.copy()
    old_grid = np.zeros_like(input_grid)
    previous_grid = np.zeros_like(input_grid)

    l_x = grid.shape[0]
    l_y = grid.shape[1]

    for i in range(steps):
        previous_grid[:, :] = old_grid
        old_grid[:, :] = grid 
        for x in range(l_x):
            for y in range(l_y):
                grid[x,y] = 0.0
                if 0 < x+1 and x + 1 < l_x:
                    grid[x,y] += old_grid[x+1,y]
                if 0 < x-1 and x- 1 < l_x:
                    grid[x,y] += old_grid[x-1,y]
                if 0 < y+1 and y+1< l_y:
                    grid[x,y] += old_grid[x,y+1]
                if 0 < y-1 and y-1< l_y:
                    grid[x,y] += old_grid[x,y-1]

                grid[x,y] /= 2.0
                grid[x,y] -= previous_grid[x,y]

    return grid

N = 5
steps = 10 
input_grid = np.random.randn(N,N).astype('float32')

fdtd(input_grid, steps)




from .. ndtypes import make_fn_type, Type 


from expr import Expr 
from stmt import block_to_str

################################################################################
#
#  Constructs below here are only used in the typed representation
#
################################################################################
class TransformHistory(object):
  """
  Sequence of transforms which have been applied to a function, 
  used for caching to avoid repeating a transformation
  """
  
  def __init__(self, transforms = None):
    self.transforms = () if transforms is None else transforms 
    self.transform_set = frozenset(self.transforms) 
    self._hash = hash(self.transform_set)
    
  def __contains__(self, T):
    return T in self.transform_set
  
  def add(self, T):
    self.transforms = self.transforms + (T,)
    self.transform_set = self.transform_set.union(frozenset([T]))
    self._hash = hash(self.transform_set)
  
  @property
  def cache_key(self):
    return self.transform_set
  
  def __hash__(self):
    return self._hash
  
  def __eq__(self, other):
    return self.transform_set == other.transform_set
  
  def __ne__(self, other):
    return self.transform_set != other.transform_set 
  
  def __str__(self):
    return str(self.transforms)
  
  def copy(self):
    return TransformHistory()

class TypedFn(Expr):
  """
  The body of a TypedFn should contain Expr nodes which have been extended with
  a 'type' attribute
  """
  
  def __init__(self, name, arg_names, body, 
                input_types, return_type,
                type_env, 
                created_by = None,
                transform_history = None,  
                source_info = None):
    
    assert isinstance(name, str), "Invalid typed function name: %s" % (name,)
    self.name = name 
    
    assert isinstance(arg_names, (list, tuple)), "Invalid typed function arguments: %s" % (arg_names,)
    self.arg_names = arg_names 

    assert isinstance(input_types, (list, tuple)), "Invalid input types: %s" % (input_types,)
    self.input_types = tuple(input_types)    
    
    assert isinstance(return_type, Type), "Invalid return type: %s" % (return_type,)
    self.return_type = return_type 
    
    assert isinstance(body, list), "Invalid body for typed function: %s" % (body,)
    self.body = body 
    
    assert isinstance(type_env, dict), "Invalid type environment: %s" % (type_env,)
    self.type_env = type_env 

    self.type = make_fn_type(self.input_types, self.return_type)
    
    self.created_by = created_by 
    
    if transform_history is None: 
      transform_history = TransformHistory()
    self.transform_history = transform_history
    
    self.source_info = source_info 


  @property 
  def cache_key(self):
    return self.name, self.created_by, self.transform_history
  
  @property
  def version(self):
    return self.transform_history.cache_key 
  
  def __repr__(self):
    arg_strings = []

    for name in self.arg_names:
      arg_strings.append("%s : %s" % (name, self.type_env.get(name)))

    return "function %s(%s) => %s:%s" % \
           (self.name, ", ".join(arg_strings),
            self.return_type,
            block_to_str(self.body))

  def __str__(self):
    return repr(self)
    # return "TypedFn(%s : %s => %s)" % (self.name, self.input_types, self.return_type)

  def __hash__(self):
    return hash(self.name) + hash(self.created_by)

  def children(self):
    return ()



import numpy as np 
from parakeet.testing_helpers import expect, run_local_tests


def test_sum_int():
  x = np.array([1,2,3])
  expect(sum, [x], sum(x))

def test_sum_float():
  x = np.array([1.0,2.0,3.0])
  expect(sum, [x], sum(x))

def sum_bool():
  x = np.array([True, False, True])
  expect(sum, [x], sum(x))

if __name__ == '__main__':
  run_local_tests()



import parakeet
import numpy as np 


def dist(x,y):
  return ((x-y)**2).sum()

def kmeans_comprehensions(X, k, niters = 10):
  C = X[:k, :]
  for _ in xrange(niters):
    A = np.array([np.argmin([dist(x,c) for c in C]) for x in X])
    C = np.array([np.mean(X[A == i, :], axis = 0) for i in xrange(k)])
  return C

def kmeans_loops(X, k, niters = 10):
  C = X[:k, :]
  n,ndims = X.shape
  A = np.zeros(n, dtype=int)
  for _ in xrange(niters):
    # assign data points to nearest centroid
    for i in xrange(n):
      x = X[i,:]
      min_dist = dist(x, C[0, :]) 
      min_idx = 0 
      for cidx in xrange(1,k):
        centroid = C[cidx,:]
        curr_dist = 0.0
        for xidx in xrange(ndims):
          curr_dist += (x[xidx] - centroid[xidx])**2
        if curr_dist < min_dist:
          min_dist = curr_dist
          min_idx = cidx
      A[i] = min_idx
    # recompute the clusters by averaging data points 
    # assigned to them 
    for cidx in xrange(k):
      # reset centroids
      for dim_idx in xrange(ndims):
        C[cidx, dim_idx] = 0
      # add each data point only to its assigned centroid
      cluster_count = 0
      for i in xrange(n):
        if A[i] == cidx:
          C[cidx, :] += X[i, :]
          cluster_count += 1
      C[cidx, :] /= cluster_count 
  return C      

n, d = 10**4, 50
X = np.random.randn(n,d)
k = 25

from compare_perf import compare_perf

compare_perf(kmeans_comprehensions, [X, k, 5],cpython=False)

compare_perf(kmeans_loops, [X, k, 5], cpython=True)



import numpy as np 
from numpy import ndarray 

from .. import syntax 
from .. syntax.helpers import const 
from ..transforms  import Transform, Simplify, Phase, DCE 


from find_constant_values import symbolic_call

from abstract_value import ( 
   Const, Array, Struct, Tuple, 
   specialization_const, abstract_tuple, abstract_array, 
   unknown, zero, one 
)

class ValueSpecializer(Transform):
  def __init__(self, abstract_inputs):
    Transform.__init__(self)
    self.abstract_inputs = abstract_inputs 
  
  def pre_apply(self, fn):
    env, _ = symbolic_call(fn, self.abstract_inputs)
    self.env = env 
  
  def lookup_expr(self, expr):
    c = expr.__class__ 
    if c is syntax.Var:
      return self.env.get(expr.name, unknown)
    else:
      return unknown 
    
  def lookup_expr_list(self, exprs):
    return [self.lookup_expr(e) for e in exprs]
  
  def transform_ParFor(self, stmt):
    fn = self.get_fn(stmt.fn)
    closure_elts = self.closure_elts(stmt.fn)
    abstract_values = self.lookup_expr_list(closure_elts)
    # all the index arguments have to be marked unknown 
    for _ in xrange(len(abstract_values), len(fn.input_types)):
      abstract_values.append(unknown)
    new_fn = specialize_abstract_values(fn, tuple(abstract_values))
    stmt.fn = self.closure(new_fn, closure_elts)
    return stmt 
    
  def transform_Var(self, expr):
    if expr.name in self.env:
      abstract_value = self.env[expr.name]
      if abstract_value.__class__ is Const:
        expr.value = expr.type.convert_python_value(abstract_value.value)
    return expr
  
  def transform_lhs(self, lhs):
    return lhs
  
def has_unit_stride(abstract_value):
  c = abstract_value.__class__
  
  if c is Const:
    return abstract_value.value == 1
  elif c is Array:
    return has_unit_stride(abstract_value.strides)
  elif c is Tuple:
    return any(has_unit_stride(elt) 
               for elt in abstract_value.elts)
  elif c is Struct:
    return any(has_unit_stride(field_val) 
               for field_val 
               in abstract_value.fields.itervalues())
  else:
    return False


def from_python(python_value):
  t = type(python_value)
  if t is ndarray:
    elt_size = python_value.dtype.itemsize 
    strides = []
    for s in python_value.strides:
      strides.append(specialization_const(s/elt_size)) 
    return abstract_array(strides)
  elif t is tuple:
    return abstract_tuple(from_python_list(python_value))
  elif python_value == 0:
    return zero 
  elif python_value == 1:
    return one 
  else:
    return unknown 
    
  
def from_python_list(python_values):
  return tuple([from_python(v) for v in python_values]) 

_cache = {}
def specialize_abstract_values(fn, abstract_values):
  key = (fn.cache_key, abstract_values)
  if key in _cache:
    return _cache[key]
  if any(has_unit_stride(v) for v in abstract_values):
    specializer = ValueSpecializer(abstract_values)
    transforms = Phase([specializer, Simplify, DCE],
                        memoize = False, 
                        copy = True, 
                        name = "StrideSpecialization for %s" % (abstract_values,), 
                        recursive = False)
    new_fn = transforms.apply(fn)
  else:
    new_fn = fn
  _cache[key] = new_fn
  return new_fn


def specialize(fn, python_values):
  return specialize_abstract_values(fn,  from_python_list(python_values))
  


from parakeet import jit 
from parakeet.testing_helpers import run_local_tests, expect

@jit 
def mod(x,y):
    return x % y

def run_mod(x, y):
  expect(mod, [x,y], x%y)

def test_mod_positive():
  run_mod(0,4)
  run_mod(4,4)
  run_mod(4,5)
  run_mod(5,4)
  run_mod(1,10000001)
  run_mod(10000001,1)

def test_mod_1st_negative():
  run_mod(-4,4)
  run_mod(-4,5)
  run_mod(-5,4)
  run_mod(-1,10000001)
  run_mod(-10000001,1)

def test_mod_2nd_negative():
  run_mod(0,-4)
  run_mod(4,-4)
  run_mod(4,-5)
  run_mod(5,-4)
  run_mod(1,-10000001)
  run_mod(10000001,-1)


def test_mod_both_negative():
  run_mod(-4,-4)
  run_mod(-4,-5)
  run_mod(-5,-4)
  run_mod(-1,-10000001)
  run_mod(-10000001,-1)

if __name__ == '__main__':
  run_local_tests()



import distutils
import os 

from . system_info import include_dirs, windows, mac_os 
from . import config 

def get_opt_flags():
  opt_flags = [config.opt_level] 
  if config.sse2:
    opt_flags.append('-msse2')
  if config.fast_math:
    opt_flags.append('-ffast-math')
  return opt_flags 

def get_compiler_flags(extra_flags = [], compiler_flag_prefix = None):
  compiler_flags = ['-I%s' % path for path in include_dirs]
  
  def add_flag(flag):
    if compiler_flag_prefix is not None:
      compiler_flags.append(compiler_flag_prefix)
    compiler_flags.append(flag)

  add_flag('-fPIC')
  
  if config.debug:
    # nvcc understands debug mode flags
    compiler_flags.extend(['-g', '-O0'])
  else:
    for flag in get_opt_flags():
      add_flag(flag)

  if not config.pure_c: 
    add_flag('-fpermissive')

  for flag in extra_flags:
    add_flag(flag)
    
  return compiler_flags   


def get_linker_flags(extra_flags = [], linker_flag_prefix = None):
  # for whatever reason nvcc is OK with the -shared linker flag
  # but not with the -fPIC compiler flag 
  linker_flags = ['-shared']
  
  def add_flag(flag):
    if linker_flag_prefix is not None:
      linker_flags.append(linker_flag_prefix)
    linker_flags.append(flag)
    
  add_flag('-lm')
  if windows:
    # crazy stupid hack for exposing Python symbols
    # even though we're compiling a shared library, why does Windows care?
    # why does windows even exist? 
    inc_dir = distutils.sysconfig.get_python_inc()
    base = os.path.split(inc_dir)[0]
    lib_dir = base + "\libs"
    add_flag('-L' + lib_dir)
    libname = 'python' + distutils.sysconfig.get_python_version().replace('.', '')
    add_flag('-l' + libname)
    
  if mac_os:
    add_flag("-Wl,-undefined")
    add_flag("-Wl,dynamic_lookup")
    
  for flag in extra_flags:
    add_flag(flag)
  return linker_flags 




import numpy as np 
import parakeet 
from parakeet.testing_helpers import expect, run_local_tests, eq, expect_allpairs 

int_mat = np.reshape(np.arange(9), (3,3))
float_mat = np.sqrt(int_mat)
bool_mat = int_mat % 2
matrices = [int_mat, float_mat, bool_mat]
vecs = [m[0] for m in matrices]

def dot(x,y):
  return sum(x*y)

def test_dot():
  expect_allpairs(dot, np.dot, vecs)

def adverb_matmult(X,Y):
  return parakeet.allpairs(dot, X, Y)


def test_adverb_matmult():
    expect_allpairs(adverb_matmult, lambda x, y: np.dot(x, y.T), matrices)

def allpairs_elt_diff(x,y):
    return parakeet.allpairs(lambda xi,yi: xi - yi, x, y)

def test_allpairs_elt_diff():
    def python_impl(x,y):
      nx = len(x)
      ny = len(y)
      result = np.zeros(shape = (nx,ny), dtype=(x[0]-y[0]).dtype)
      for i in xrange(nx):
          for j in xrange(ny):
              result[i,j] = x[i] - y[j]
      return result 
    expect_allpairs(allpairs_elt_diff, python_impl, vecs)


if __name__ == '__main__':
  run_local_tests()



from loop_transform import LoopTransform
from transform import Transform

class ExtractParallelLoops(Transform):
  
  def __init__(self):
    Transform.__init__(self)
    
  def enter_loop(self, var):
    



# Authors: Nathan Faggian, Stefan van der Walt, Aron Ahmadia, Olivier Grisel
# https://github.com/stefanv/growcut_py

import numpy as np

def growcut_python(image, state, state_next, window_radius):
    changes = 0
    height = image.shape[0]
    width = image.shape[1]
    for j in xrange(width):
        for i in xrange(height):
            winning_colony = state[i, j, 0]
            defense_strength = state[i, j, 1]
            for jj in xrange(max(j-window_radius,0), min(j+window_radius+1, width)):
                for ii in xrange(max(i-window_radius, 0), min(i+window_radius+1, height)):
                    if ii != i or jj != j:
                        d = image[i, j, :] - image[ii, jj, :]
                        s = np.sum(d**2) 
                        gval = 1.0 - np.sqrt(s) / np.sqrt(3)
                        attack_strength = gval * state[ii, jj, 1]
                        if attack_strength > defense_strength:
                            defense_strength = attack_strength
                            winning_colony = state[ii, jj, 0]
                            changes += 1
            state_next[i, j, 0] = winning_colony
            state_next[i, j, 1] = defense_strength
    return changes
    
N = 50
dtype = np.double
image = np.zeros((N, N, 3), dtype=dtype)
state = np.zeros((N, N, 2), dtype=dtype)
state_next = np.empty_like(state)

# colony 1 is strength 1 at position 0,0
# colony 0 is strength 0 at all other positions
state[0, 0, 0] = 1
state[0, 0, 1] = 1

window_radius = 10

from compare_perf import compare_perf 

compare_perf(growcut_python, [image, state, state_next, window_radius])



import sys
import time 
import numpy as np

from nose.tools import nottest

from dsltools.testing_helpers import eq, expect_eq, run_local_tests
  
from . import (type_conv, type_inference, config, 
              specialize, translate_function_value,
              find_broken_transform,
              run_untyped_fn, 
              run_typed_fn, 
              run_python_fn)




def _copy(x):
  if isinstance(x, np.ndarray):
    return x.copy()
  else:
    return x

def _copy_list(xs):
  return [_copy(x) for x in xs]


def expect(fn, args, expected, msg = None, valid_types = None):
  """
  Helper function used for testing, assert that Parakeet evaluates given code to
  the correct result
  """
  if hasattr(expected, 'dtype') and expected.dtype == 'float16':
    expected = expected.astype('float32')

  untyped_fn = translate_function_value(fn)
  
  available_backends = ['interp', 'c', 'openmp']

  #import cuda_backend 
  #if cuda_backend.device_info.has_gpu():
  #  available_backends.append('cuda')
     

  for backend in available_backends: #available_backends:
    try: 
      result = run_untyped_fn(untyped_fn, _copy_list(args), backend = backend)
      
    except: 
      if config.testing_find_broken_transform: 
        find_broken_transform(fn, args, expected)
      raise
   
    label = "backend=%s, inputs = %s" % (backend, ", ".join(str(arg) for arg in args))

    if msg is not None:
      label += ": " + str(msg)
    
    try: 
      #if hasattr(result, 'flags'):
      #  print result.flags
      expect_eq(result, expected, label)
    except: 
      if config.testing_find_broken_transform: 
        find_broken_transform(fn, args, expected)
      raise 
    
    if valid_types is not None:
      if not isinstance(valid_types, (tuple, list)):
        valid_types = [valid_types]
      assert type(result) in valid_types, \
        "Expected result to have type in %s but got %s" % (valid_types, type(result))
  
def expect_each(parakeet_fn, python_fn, inputs):
  for x in inputs:
    expect(parakeet_fn, [x], python_fn(x))

def expect_allpairs(parakeet_fn, python_fn, inputs):
  for x in inputs:
    for y in inputs:
      expect(parakeet_fn, [x,y], python_fn(x,y))

def return_type(fn, input_types):
  untyped_fundef = translate_function_value(fn)
  closure_args = untyped_fundef.python_nonlocals()
  closure_arg_types = map(type_conv.typeof, closure_args)
  return type_inference.infer_return_type(untyped_fundef,
                                          closure_arg_types + input_types)

def expect_type(fn, input_types, output_type):
  actual = return_type(fn, input_types)
  assert actual == output_type, "Expected type %s, actual %s" % \
                                (output_type, actual)

def assert_eq_arrays(numpy_result, parakeet_result, test_name = None):
  if test_name is None:
    msg = ""
  else:
    msg = "[%s] " % test_name
  assert type(numpy_result) == type(parakeet_result), \
    "%sExpected type %s but got %s" % (msg, type(numpy_result), type(parakeet_result))
  if hasattr(numpy_result, 'shape'):
    assert hasattr(parakeet_result, 'shape')
    assert numpy_result.shape == parakeet_result.shape, \
      "%sExpected shape %s but got %s" % (msg, numpy_result.shape, parakeet_result.shape)
    assert eq(numpy_result, parakeet_result), \
      "%sExpected value %s but got %s" % (msg, numpy_result, parakeet_result)
    
@nottest
def timed_test(parakeet_fn, parakeet_args, python_fn, 
               python_args = None, min_speedup = None):
  if python_args is None:
    python_args = parakeet_args

  start = time.time()
  _ = parakeet_fn(*parakeet_args)
  parakeet_time_with_comp = time.time() - start 

  start = time.time()
  py_result = python_fn(*python_args)
  py_time = time.time() - start 

  start = time.time()
  parakeet_result = parakeet_fn(*parakeet_args)
  parakeet_time_no_comp = time.time() - start 

  print "Parakeet time (with compilation):", parakeet_time_with_comp
  print "Parakeet time (without compilation):", parakeet_time_no_comp
  print "Python time:", py_time 

  assert eq(parakeet_result, py_result), \
    "Expected %s but got %s" % (py_result, parakeet_result)
  if min_speedup is not None:
    assert py_time / parakeet_time_no_comp > min_speedup, \
        "Parakeet too slow: %.2f slowdown" % (parakeet_time_no_comp/py_time)



from .. syntax import Assign, Index, Var

from loop_transform import LoopTransform

class RedundantLoadElimination(LoopTransform):
  def transform_block(self, stmts):
    stmts = LoopTransform.transform_block(self, stmts) 
    if self.is_simple_block(stmts, allow_branches = False):
      reads, writes = self.collect_memory_accesses(stmts)
      safe_arrays = set([])
      for name in reads:
        # if any alias of this array gets written to, consider it unsafe 
        aliases = self.may_alias.get(name, set([]))
        aliases.add(name)
        unsafe = any(alias in writes for alias in aliases) 
        if not unsafe:
          safe_arrays.add(name)
      available_expressions = {}
      new_stmts = []
      for stmt in stmts:
        if stmt.__class__ is Assign:
          if stmt.rhs.__class__ is Index and \
             stmt.rhs.value.__class__ is Var and \
             stmt.rhs.value.name in safe_arrays:
            key = (stmt.rhs.value.name, stmt.rhs.index)
            if key in available_expressions:
              stmt.rhs = available_expressions[key]
            elif stmt.lhs.__class__ is Var:
              available_expressions[key] = stmt.lhs
            else:
              temp = self.fresh_var(stmt.rhs.type,  "load")
              new_stmts.append(Assign(temp, stmt.rhs))
              stmt.rhs = temp
              available_expressions[key] = temp
          new_stmts.append(stmt)
      return new_stmts
    else:
      return stmts
        
        


import numpy as np 
from parakeet import jit, testing_helpers


alpha = 0.5
beta = 0.3
x = np.array([1,2,3])
y = np.tanh(x) * alpha + beta

@jit
def numpy_expressions(x, alpha = 0.5, beta = 0.3):
  return np.tanh(x) * alpha + beta 

def test_numpy_expressions():
  res = numpy_expressions(x)
  assert np.allclose(res, y), "Expected %s but got %s" % (y, res)

@jit 
def loopy(x, alpha = 0.5, beta = 0.3):
  y = np.empty_like(x, dtype=float)
  for i in xrange(len(x)):
    y[i] = np.tanh(x[i]) * alpha + beta
  return y

def test_loopy():
  res = loopy(x)
  assert np.allclose(res, y), "Expected %s but got %s" % (y, res)

@jit
def comprehension(x, alpha = 0.5, beta = 0.3):
  return np.array([np.tanh(xi)*alpha + beta for xi in x])

def test_comprehensions():
  res = comprehension(x)
  assert np.allclose(res, y), "Expected %s but got %s" % (y, res)

if __name__ == "__main__":
  testing_helpers.run_local_tests()
  



from dsltools import NestedBlocks

from .. import syntax, names 
from ..ndtypes import (make_tuple_type, 
                       Int32, Int64, SliceT, TupleT, ScalarT, StructT, NoneT, 
                       ArrayT, FnT, ClosureT, NoneType) 
from ..syntax import (Assign, Return, If,    
                      ArrayView, Attribute, Cast, Const, Closure,  Comment, Expr, 
                      Index, PrimCall,   Struct, Slice, Tuple, TupleProj, TypedFn, Var, 
                      AllocArray, ArrayExpr, Adverb, Select)

from ..syntax.helpers import (const_bool, const_int, get_types,  
                              one_i64, zero, none)

class CoreBuilder(object):
  none = syntax.none
  null_slice = syntax.slice_none
  
  def __init__(self, type_env = None, blocks = None):
    if type_env is None:
      type_env = {}
    self.type_env = type_env 
    if blocks is None: 
      blocks = NestedBlocks()
      
    self.blocks = blocks 

    # cut down the number of created nodes by
    # remembering which tuple variables we've created
    # and looking up their elements directly
    self.tuple_elt_cache = {}

  def insert_stmt(self, stmt):
    self.blocks.append_to_current(stmt)

  def append(self, stmt):
    self.blocks.append_to_current(stmt)
  
  def comment(self, text):
    self.blocks.append(Comment(text))
  
  def assign(self, lhs, rhs):
    self.insert_stmt(Assign(lhs, rhs))
  
  def return_(self, value):
    self.blocks += [Return(value)]  
  
  def return_none(self):
    self.return_(none)
  
  def return_tuple(self, values):
    self.return_(self.tuple(values, name = None))
    
  def fresh_var(self, t, prefix = "temp"):
    assert prefix is not None
    assert t is not None, "Type required for new variable %s" % prefix
    prefix = names.original(prefix)
    ssa_id = names.fresh(prefix)
    self.type_env[ssa_id] = t
    return Var(ssa_id, type = t)


  def if_(self, cond, true_thunk, false_thunk, result_vars = ()):
    """
    Args:
      cond - the condition expression 
      true_thunk : given some variables, construct the true branch
      false_thunk : given other variables, construct the false branch
      result_vars : the merged version of the values from the two branches
    """
    merge = {}
    assert all(expr.__class__ is Var for expr in result_vars)
    left_vars = []
    right_vars = []
    for result_var in result_vars:
      name = result_var.name 
      t = result_var.type
      left_var = self.fresh_var(t, prefix = "if_true_" + names.original(name) )
      left_vars.append(left_var)
      right_var = self.fresh_var(t, prefix = "if_false_" + names.original(name) )
      right_vars.append(right_var)
      merge[result_var.name] = (left_var, right_var)
    
    self.blocks.push()
    true_thunk(*left_vars)
    true_block = self.blocks.pop()
    
    self.blocks.push()
    false_thunk(*right_vars)
    false_block = self.blocks.pop()
     
    stmt = If(cond, true_block, false_block, merge = merge)
    self.insert_stmt(stmt)
    return stmt 

  def fresh_i32(self, prefix = "temp"):
    return self.fresh_var(Int32, prefix)

  def fresh_i64(self, prefix = "temp"):
    return self.fresh_var(Int64, prefix)

  
  def temp_name(self, expr):
    c = expr.__class__
    if c is PrimCall:
      return expr.prim.name
    elif c is Attribute:
      if expr.value.__class__ is Var:
        return names.original(expr.value.name) + "_" + expr.name
      else:
        return expr.name
    elif c is Attribute:
      if expr.value.__class__ is Var:
        return "%s_%s" % (expr.value.name, expr.name)
      else:
        return expr.name
    elif c is Index:
      idx_t = expr.index.type
      if isinstance(idx_t, SliceT) or \
         (isinstance(idx_t, TupleT) and \
          any(isinstance(elt_t, SliceT) for elt_t in idx_t.elt_types)):
        return "slice"
      else:
        return "elt"
    elif c is TupleProj:
      if expr.tuple.__class__ is Var:
        original = names.original(expr.tuple.name)
        return "%s%d_elt%d" % (original, names.versions[original], expr.index)
      else:
        return "tuple_elt%d" % expr.index
    elif c is Var:
      return names.refresh(expr.name) 
    else:
      return "temp"

  def is_simple(self, expr):
    c = expr.__class__
    return c is Var or \
           c is Const or \
           (c is Tuple and len(expr.elts) == 0) or \
           (c is Struct and len(expr.args) == 0) or \
           (c is Closure and len(expr.args) == 0)


  def is_pure(self, expr):
    c = expr.__class__
    return c is Var or \
           c is Const or \
           c is Tuple or \
           c is Struct or \
           c is Closure or \
           c is AllocArray or \
           c is ArrayView or \
           c is PrimCall or \
           isinstance(expr, ArrayExpr) or \
           isinstance(expr, Adverb)
           
    



  def assign_name(self, expr, name = None):
    #if self.is_simple(expr):
    #  return expr
    if name is None:
      name = self.temp_name(expr)
    
    var = self.fresh_var(expr.type, names.refresh(name))
    self.assign(var, expr)
    return var

  def int(self, x, name = None):
    if name is None: 
      return const_int(x)
    else:
      return self.assign_name(const_int(x), name)

  def bool(self, x, name = None):
    if name is None: 
      return const_bool(x)
    else:
      return self.assign_name(const_bool(x), name)

  def zero(self, t = Int32, name = None):
    if name is None:
      return zero(t)
    else:
      return self.assign_name(zero(t), name)

  def zero_i32(self, name = None):
      return self.zero(t = Int32, name = name)

  def zero_i64(self, name = None):
    return self.zero(t = Int64, name = name)

  def cast(self, expr, t):
    if expr.type == t:
      return expr
    assert isinstance(t, ScalarT), \
        "Can't cast %s : %s to non-scalar type %s" % (expr, expr.type, t)
    return self.assign_name(Cast(expr, type = t), "cast_%s" % t)
  


  def attr(self, obj, field, name = None, temp = True):
    if temp and name is None:
      name = field
    obj_t = obj.type
    c = obj.__class__ 
    if c is Struct:
      pos = obj_t.field_pos(name)
      result =  obj.args[pos]
    elif c in (ArrayView, Slice):
      result = getattr(obj, field)
    else:
      assert isinstance(obj_t, StructT), \
        "Can't get attribute '%s' from type %s" % (field, obj_t)
      field_t = obj.type.field_type(field)
      if field_t == NoneType:
        result = none 
      else:
        result = Attribute(obj, field, type = field_t)
    if name and result.__class__ not in (Var, Const):
      return self.assign_name(result, name)
    else:
      return result

  def is_none(self, x):
    return x is None or (isinstance(x, Expr) and x.type.__class__ is NoneT)


  def tuple(self, elts, name = None, explicit_struct = False):
    if not isinstance(elts, (list, tuple)):
      elts = [elts]
    tuple_t = make_tuple_type(get_types(elts))
    if explicit_struct:
      tuple_expr = Struct(elts, type = tuple_t)
    else:
      tuple_expr = Tuple(elts, type = tuple_t)
    if name:
      result_var = self.assign_name(tuple_expr, name)
      # cache the simple elements so we can look them up directly
      for (i, elt) in enumerate(elts):
        if self.is_simple(elt):
          self.tuple_elt_cache[(result_var.name, i)] = elt
      return result_var
    else:
      return tuple_expr

  def is_tuple(self, x):
    try:
      return x.type.__class__ is TupleT
    except:
      return False
    
  def is_array(self, x):
    return x.type.__class__ is  ArrayT
  
  def is_fn(self, x):
    return x.__class__ in (TypedFn, Closure) or isinstance(x.type, (FnT, ClosureT)) 
  
  def num_tuple_elts(self, x):
    assert self.is_tuple(x)
    return len(x.type.elt_types)
  
  def concat_tuples(self, x, y, name = "concat_tuple"):
    
    x_is_tuple = self.is_tuple(x)
    y_is_tuple = self.is_tuple(y)
    
    if x_is_tuple and y_is_tuple:
      if self.num_tuple_elts(x) == 0:
        return y 
      elif self.num_tuple_elts(y) == 0:
        return x 
      
    if x_is_tuple:
      x_elts = self.tuple_elts(x)
    else:
      x_elts = (x,)
    if y_is_tuple:
      y_elts = self.tuple_elts(y)
    else:
      y_elts = (y,)

    elts = []
    elts.extend(x_elts)
    elts.extend(y_elts)
    return self.tuple(elts, name = name)

  def tuple_proj(self, tup, idx, explicit_struct = False):
    assert isinstance(idx, (int, long))
    if isinstance(tup, Tuple):
      return tup.elts[idx]
    elif isinstance(tup, tuple):
      return tup[idx]
    elif tup.__class__ is Var and (tup.name, idx) in self.tuple_elt_cache:
      return self.tuple_elt_cache[(tup.name, idx)]
    elif explicit_struct:
      return Attribute(tup, "elt%d" % idx, type = tup.type.elt_types[idx])
    else:
      return TupleProj(tup, idx, type = tup.type.elt_types[idx])

  def tuple_elts(self, tup, explicit_struct = False):
    if not isinstance(tup.type, TupleT):
      return [tup]
    nelts = len(tup.type.elt_types)
    return tuple([self.tuple_proj(tup, i, explicit_struct = explicit_struct)
                  for i in xrange(nelts)])

  def prod(self, elts, name = None):
    if self.is_tuple(elts):
      elts = self.tuple_elts(elts)
    if len(elts) == 0:
      return one_i64
    else:
      result = elts[0]
      for e in elts[1:]:
        result = self.mul(result, e, name = name)
      return result
    
  def select(self, cond, true_value, false_value, name = None):
    assert true_value.type == false_value.type
    expr = Select(cond, true_value, false_value, type = true_value.type)
    if name is None:
      return expr
    else:
      return self.assign_name(expr, name)


import numpy as np
from parakeet.testing_helpers import run_local_tests, expect_eq 
import parakeet

def mk_scalar_add(t):
  f, b, (x,y) = parakeet.build_fn([t,t], t)
  b.return_(b.add(x,y))
  return f

def mk_vec_add(array_t):
  elt_t = parakeet.elt_type(array_t)
  add_fn = mk_scalar_add(elt_t)
  fn, builder, inputs = parakeet.build_fn([array_t, array_t, array_t])
  x,y,z = inputs
  n = builder.len(x)
  def loop_body(idx):
    elt_x = builder.index(x, idx)
    elt_y = builder.index(y, idx)
    builder.setidx(z, idx, builder.call(add_fn, (elt_x, elt_y)))
  builder.loop(0, n, loop_body)
  builder.return_(builder.none)
  return fn 


def vec_add(x,y):
  assert x.dtype == y.dtype 
  assert len(x.shape) == len(y.shape) == 1 
  z = np.zeros_like(x)
  array_t = parakeet.typeof(x)
  fn = mk_vec_add(array_t)
  parakeet.run_typed_fn(fn, (x,y,z))
  return z 
  
def test_vec_add(): 
  xs,ys = np.array([1,2,3]), np.array([10,20,30])
  zs = vec_add(xs, ys)
  expected = xs + ys 
  expect_eq(zs, expected)
  

if __name__ == '__main__': 
  run_local_tests()  



from .. import names 
from .. syntax import (TypedFn, Var, Const, Attribute, Index, PrimCall, 
                       If, Assign, While, ExprStmt, Return, ForLoop, ParFor,  
                       Slice, Struct, Tuple, TupleProj, Cast, Alloc, Closure, 
                       Map, Reduce, Scan, IndexMap, IndexReduce, IndexScan, 
                       UntypedFn )      
from transform import Transform 
from parakeet.syntax.delay_until_typed import DelayUntilTyped

class CloneFunction(Transform):
  """
  Copy all the objects in the AST of a function
  """
  def __init__(self, parent_transform = None, 
                      rename = False):
    Transform.__init__(self)
    self.rename = rename 
    # self.recursive = recursive
    self.parent_transform = parent_transform 
     
  def transform_Var(self, expr):
    return Var(expr.name, type = expr.type)
  
  def transform_expr(self, expr):
    new_expr = self._transform_expr(expr)
    new_expr.type = expr.type 
    new_expr.source_info = expr.source_info 
    return new_expr 
  
  def _transform_expr(self, expr):
    c = expr.__class__
    if c is Var:
      return Var(expr.name)
      
    elif c is Const:
      return Const(expr.value)
    
    elif c is Tuple:
      new_elts = tuple(self.transform_expr(elt) for elt in expr.elts)
      return Tuple(elts = new_elts)
    
    elif c is Attribute:
      value = self.transform_expr(expr.value)
      return Attribute(value, expr.name)
    
    elif c is Index:
      value = self.transform_expr(expr.value)
      index = self.transform_expr(expr.index)
      return Index(value, index)
    
    elif c is PrimCall:
      args = tuple(self.transform_expr(elt) for elt in expr.args)
      return PrimCall(expr.prim, args)
    
    elif c is TypedFn:
      #if self.recursive:
      #  cloner = CloneFunction(rename = self.rename, recursive = True)
      #  return cloner.apply(expr)
      #else:
      return expr 
    elif c is UntypedFn:
      return expr 
    elif c is Closure: 
      args = self.transform_expr_tuple(expr.args)
      return Closure(fn = expr.fn, args = args)
    
    elif c is Slice: 
      start = self.transform_if_expr(expr.start)
      stop = self.transform_if_expr(expr.stop)
      step = self.transform_if_expr(expr.step)
      return Slice(start, stop, step, type = expr.type)
    
    elif c is Struct: 
      new_args = self.transform_expr_list(expr.args)
      return Struct(args = new_args, type = expr.type)
    
    elif c is TupleProj: 
      new_tuple = self.transform_expr(expr.tuple)
      return TupleProj(new_tuple, expr.index, type = expr.type)
    
    elif c is Cast:
      return Cast(self.transform_expr(expr.value), expr.type)
     
    elif c is Alloc:
      return Alloc(count = self.transform_expr(expr.count), 
                   elt_type = expr.elt_type)
    elif c is Map:
      return Map(fn = self.transform_expr(expr.fn), 
                  args = self.transform_expr_list(expr.args), 
                  axis = self.transform_if_expr(expr.axis), 
                 )
    elif c is Reduce: 
      return Reduce(fn = self.transform_expr(expr.fn), 
                    combine = self.transform_expr(expr.combine), 
                    args = self.transform_expr_list(expr.args), 
                    axis = self.transform_if_expr(expr.axis), 
                    init = self.transform_if_expr(expr.init), 
                  )
     
    else:
      args = {}  
      for k,v in expr.__dict__.iteritems():
        args[k] = self.transform_if_expr(v)
      if c is DelayUntilTyped and 'type' in args:
        del args['type']
      return c(**args)
  
  def transform_Assign(self, stmt):

    new_lhs = self.transform_expr(stmt.lhs)
    new_rhs = self.transform_expr(stmt.rhs)
    return Assign(new_lhs, new_rhs)

  def transform_ExprStmt(self, stmt):
    return ExprStmt(self.transform_expr(stmt.value))

  def transform_Return(self, stmt):
    res = Return(self.transform_expr(stmt.value))
    return res 
  
  def transform_If(self, stmt):
    new_true = self.transform_block(stmt.true)
    new_false = self.transform_block(stmt.false)
    new_merge = self.transform_merge(stmt.merge)
    new_cond = self.transform_expr(stmt.cond)
    return If(new_cond, new_true, new_false, new_merge)
  
  def transform_While(self, stmt):
    new_body = self.transform_block(stmt.body)
    new_merge = self.transform_merge(stmt.merge)
    new_cond = self.transform_expr(stmt.cond)
    return While(new_cond, new_body, new_merge)
  
  def transform_ForLoop(self, stmt):
    new_var = self.transform_expr(stmt.var)
    new_start = self.transform_expr(stmt.start)
    new_stop = self.transform_expr(stmt.stop)
    new_step = self.transform_expr(stmt.step)
    new_body = self.transform_block(stmt.body)
    new_merge = self.transform_merge(stmt.merge)
    return ForLoop(new_var, new_start, new_stop, new_step, new_body, new_merge)  
  
  def transform_ParFor(self, stmt):
    new_bounds = self.transform_expr(stmt.bounds)
    new_fn = self.transform_expr(stmt.fn)
    return ParFor(fn = new_fn, bounds = new_bounds)
  
  def pre_apply(self, old_fn):
    new_fundef_args = old_fn.__dict__.copy()
    del new_fundef_args['type']
    # new_fundef_args = dict([(m, getattr(old_fn, m)) for m in old_fn._members])
    # create a fresh function with a distinct name and the
    # transformed body and type environment
    if self.rename: 
      new_fundef_args['name'] = names.refresh(old_fn.name)
    else:
      new_fundef_args['name'] = old_fn.name
       
    new_fundef_args['type_env'] = old_fn.type_env.copy()
    new_fundef_args['transform_history'] = old_fn.transform_history.copy()
    new_fundef_args['created_by'] = self.parent_transform
    # don't need to set a new body block since we're assuming 
    # that transform_block will at least allocate a new list 
    new_fundef = TypedFn(**new_fundef_args)
    return new_fundef 




from ..ndtypes import ScalarT, PtrT, NoneT, TupleT, FnT, ClosureT 
from .. syntax import Adverb, ParFor, Closure, UntypedFn, TypedFn, ArrayExpr  

from syntax_visitor import SyntaxVisitor

def memoize(analyzer):
  _cache = {}
  def memoized(fn_arg): 
    key = analyzer, fn_arg.cache_key
    if key in _cache:
      return _cache[key]
    result = analyzer(fn_arg)
    _cache[key] = result 
    return result 
  return memoized  
  

@memoize 
def contains_structs(fn):
  for t in fn.type_env.itervalues():
    if not isinstance(t, (ScalarT, PtrT, NoneT)):
      return True
  return False

class Yes(Exception):
  pass 

class ContainsSlices(SyntaxVisitor):
  def visit_Index(self, expr):
    if isinstance(expr.index.type, ScalarT):
      idx_types = [expr.index.type]
    elif isinstance(expr.index.type, TupleT):
      idx_types = expr.index.type.elt_types
    else: 
      raise Yes()
    if not all(isinstance(t, ScalarT) for t in idx_types):
      raise Yes()

@memoize 
def contains_slices(fn):
  try:
    ContainsSlices().visit_fn(fn)
    return False 
  except Yes:
    return True
  
class ContainsLoops(SyntaxVisitor):
  def visit_ForLoop(self, expr):
    raise Yes()
  
  def visit_While(self, expr):
    raise Yes()

def contains_loops(fn):
  try:
    ContainsLoops().visit_fn(fn)
    return False 
  except Yes:
    return True
  

class ContainsCalls(SyntaxVisitor):
  def visit_Call(self, expr):
    raise Yes()
  
  def visit_TypedFn(self, expr):
    if contains_calls(expr):
      raise Yes()
  
@memoize 
def contains_calls(fn):
  try: 
    ContainsCalls().visit_fn(fn)
    return False 
  except:
    return True 

class ContainsAdverbs(SyntaxVisitor):
  def visit_expr(self, expr):
    if isinstance(expr, Adverb):
      raise Yes()
    SyntaxVisitor.visit_expr(self, expr)
  
  def visit_TypedFn(self, expr):
    if contains_adverbs(expr):
      raise Yes()
  
@memoize 
def contains_adverbs(fn):
  try:
    ContainsAdverbs().visit_fn(fn)
    return False
  except Yes:
    return True

class ContainsParFor(SyntaxVisitor):
  def visit_stmt(self, stmt):
    if stmt.__class__ is ParFor:
      raise Yes()
    SyntaxVisitor.visit_stmt(self, stmt)
  
  
  def visit_TypedFn(self, expr):
    if contains_parfor(expr):
      raise Yes()

@memoize 
def contains_parfor(fn):
  try:
    ContainsParFor().visit_fn(fn)
    return False 
  except Yes:
    return True 

class ContainsFunctions(SyntaxVisitor):
  def visit_expr(self, expr):
    if isinstance(expr, (UntypedFn, TypedFn, Closure)) or isinstance(expr.type, (FnT, ClosureT)):
      raise Yes()
    SyntaxVisitor.visit_expr(expr)

@memoize 
def contains_functions(fn):
  try:
    ContainsFunctions().visit_fn(fn)
    return False 
  except Yes:
    return True 

class ContainsAlloc(SyntaxVisitor):
  def visit_Alloc(self, _):
    raise Yes()
  
  def visit_AllocArray(self, _):
    raise Yes()
  
  def visit_TypedFn(self, fn):
    if contains_alloc(fn):
      raise Yes()  
      
@memoize 
def contains_alloc(fn):
  try:
    ContainsAlloc().visit_fn(fn) 
    return False 
  except Yes:
    return True

class ContainsArrayOperators(SyntaxVisitor):
  def visit_expr(self, expr):
    """
    We're assuming that array expressions only occur on RHS of assignments
    and aren't nested. It's up to Simplify to de-nest them
    """
    if isinstance(expr, ArrayExpr):
      raise Yes()
  
@memoize 
def contains_array_operators(fn):
  try:
    ContainsArrayOperators().visit_fn(fn) 
    return False 
  except Yes:
    return True        



from ..ndtypes import FnT, ClosureT, Type, make_closure_type, NoneType 
from ..syntax import UntypedFn, TypedFn, Var, Call, Closure, ClosureElt, Return, FormalArgs   
from ..syntax.helpers import get_types, zero_i64, none 
from ..syntax.adverb_helpers import max_rank

from core_builder import CoreBuilder
from parakeet.syntax.stmt import ExprStmt

class CallBuilder(CoreBuilder):
  
  """
  Builder for all things related to calling functions
  """
  
  def closure_elt(self, clos, idx, name = None):
    assert isinstance(idx, (int, long))

    if isinstance(clos, Closure):
      result = clos.args[idx]
    else:
      result = ClosureElt(clos, idx, type = clos.type.arg_types[idx])
    if name is None:
      return result
    return self.assign_name(result, name)
  
  def closure_elts(self, clos):
    if clos.__class__ is TypedFn:
      return ()
    return tuple(self.closure_elt(clos, i)
                 for i in xrange(len(clos.type.arg_types)))

  def get_fn(self, maybe_clos):
    if maybe_clos.__class__ is Closure:
      return maybe_clos.fn
    elif maybe_clos.type.__class__ is ClosureT:
      return maybe_clos.type.fn
    else:
      return maybe_clos

  def closure(self, maybe_fn, extra_args, name = None):
    fn = self.get_fn(maybe_fn)
    old_closure_elts = self.closure_elts(maybe_fn)
    closure_elts = old_closure_elts + tuple(extra_args)
    if len(closure_elts) == 0:
      return fn 
    closure_elt_types = [elt.type for elt in closure_elts]
    closure_t = make_closure_type(fn, closure_elt_types)
    result = Closure(fn, closure_elts, type = closure_t)
    if name:
      return self.assign_name(result, name)
    else:
      return result
  

  def return_type(self, fn):
    if isinstance(fn, TypedFn):
      return fn.return_type
    else:
      assert isinstance(fn.type, ClosureT), \
          "Unexpected fn type: %s" % fn.type
      assert isinstance(fn.type.fn, TypedFn), \
          "Unexpected fn: %s" % fn.type.fn
      return fn.type.fn.return_type


  def input_vars(self, fn):
    assert isinstance(fn, TypedFn), "Expected TypedFn, got %s" % fn 
    
    return [Var(arg_name, t) 
            for arg_name, t in 
            zip(fn.arg_names, fn.input_types)]
  
  def input_types(self, closure):
    fn = self.get_fn(closure)
    closure_args = self.closure_elts(closure)
    return fn.input_types[len(closure_args):]
  
  def invoke_type(self, closure, args):
    from .. type_inference import invoke_result_type 
    closure_t = closure.type
    arg_types = get_types(args)
    assert all(isinstance(t, Type) for t in arg_types), \
        "Invalid types: %s" % (arg_types, )
    return invoke_result_type(closure_t, arg_types)

  def is_identity_fn(self, fn):
    if fn.__class__ is TypedFn and len(fn.arg_names) == 1:
      input_name = fn.arg_names[0]
    elif fn.__class__ is UntypedFn:
      args = fn.args 
      if isinstance(args, (list, tuple)):
        input_name = args[0]
      else:
        assert isinstance(args, FormalArgs), "Unexpected args %s" % (args,)
        if args.n_args == 1 and len(args.positional) == 1:
          input_name = args.positional[0]  
        else:
          return False 
    else:
      return False 
    
    if isinstance(input_name, Var):
      input_name = input_name.name
    else:
      assert isinstance(input_name, str), "Unexpected input %s" % (input_name,)    
    if len(fn.body) == 1:
      stmt = fn.body[0]
      if stmt.__class__ is Return:
        expr = stmt.value 
        if expr.__class__ is Var:
          return expr.name == input_name
    return False 
    
  def invoke(self, fn, args, loopify = False, lower = False, name = None):
    

    if isinstance(fn, UntypedFn) or isinstance(fn.type, FnT):
      closure_args = [] 
    else: 
      assert isinstance(fn.type, ClosureT), "Unexpected function %s with type: %s" % (fn, fn.type)
      closure_args = self.closure_elts(fn)
      fn = self.get_fn(fn)
      
    
    args =  tuple(closure_args) + tuple(args)
    
    if isinstance(fn, UntypedFn):
      arg_types = get_types(args)
      from .. import type_inference
      fn = type_inference.specialize(fn, arg_types)
      
    if loopify or lower: 
      from  ..transforms import pipeline
      if loopify: 
        fn = pipeline.loopify(fn)
      if lower: 
        fn = pipeline.lowering(fn)
    
    # don't generate Call nodes for identity function 
    if self.is_identity_fn(fn):
      assert len(args) == 1
      return args[0]
    
  
    call = Call(fn, args, type = fn.return_type)

    if fn.return_type == NoneType:
      self.insert_stmt(ExprStmt(call))
      return none
    else:
      if name is None:
        name = "call_result"
      return self.assign_name(call, name)

  def call(self, fn, args, name = None):
    return self.invoke(fn, args, name = name) 

  def call_shape(self, maybe_clos, args):
    from ..shape_inference import call_shape_expr, shape_codegen
    fn = self.get_fn(maybe_clos)
    closure_args = self.closure_elts(maybe_clos)
    combined_args = tuple(closure_args) + tuple(args)
    if isinstance(fn, UntypedFn):
      # if we're given an untyped function, first specialize it
      from ..type_inference import specialize       
      fn = specialize(fn, get_types(combined_args))
      from ..transforms import pipeline 
      fn = pipeline.high_level_optimizations(fn)
    abstract_shape = call_shape_expr(fn)
    return shape_codegen.make_shape_expr(self, abstract_shape, combined_args)





from adverbs import * 
from array_properties import *
from array_constructors import *
from builtins import * 
from math import *
from numpy_misc import *  
from reductions import * 
from numpy_types import * 
from patchmap import *

import linalg 
from linalg import dot 
import random 


import itertools 
import numpy as np
import types


from frontend import ast_conversion
from ndtypes import ScalarT, StructT,  type_conv     
from syntax import (Expr, Var, Tuple, 
                    UntypedFn, TypedFn, 
                    Return, If, While, ForLoop, ParFor, ExprStmt,   
                    ActualArgs, 
                    Assign, Index, )


class ReturnValue(Exception):
  def __init__(self, value):
    self.value = value

class ClosureVal:
  def __init__(self, fn, fixed_args):
    self.fn = fn
    self.fixed_args = tuple(fixed_args)

  def __call__(self, args):
    if isinstance(args, ActualArgs):
      args = args.prepend_positional(self.fixed_args)
    else:
      args = self.fixed_args + tuple(args)
    return eval_fn(self.fn, args)
  
def rankof(x):
  return x.ndim if hasattr(x, 'ndim') else 0 

def eval(fn, actuals):
  
  result = eval_fn(fn, actuals)
  import ctypes 
  if isinstance(result, ctypes.Structure):
    return type_conv.to_python(result, fn.return_type)
  return result 
   
def eval_fn(fn, actuals):
  
  if isinstance(fn, np.dtype):
    return fn.type(*actuals)
  elif isinstance(fn, TypedFn):
    assert len(fn.arg_names) == len(actuals), \
      "Wrong number of args, expected %s but given %s" % \
      (fn.arg_names, actuals)
    env = {}

    for (k,v) in zip(fn.arg_names, actuals):
      env[k] = v
  elif isinstance(fn, UntypedFn):
    # untyped functions have a more complicated args object
    # which deals with named args, variable arity, etc..
    env = fn.args.bind(actuals)
  elif isinstance(fn, ClosureVal):
    return fn(actuals)
  else:
    return fn(*actuals)
  

  def eval_args(args):
    if isinstance(args, (list, tuple)):
      return map(eval_expr, args)
    else:
      return args.transform(eval_expr)

  def eval_if_expr(maybe_expr):
    return eval_expr(maybe_expr) if isinstance(maybe_expr, Expr) else maybe_expr
  
  def eval_expr(expr):

    if hasattr(expr, 'wrapper'):
      expr = expr.wrapper
    assert isinstance(expr, Expr), "Not an expression-- %s : %s" % \
         (expr, type(expr))
         
    def expr_Const():
      return expr.value

    def _strides(numpy_array):
      strides_bytes = numpy_array.strides
      itemsize = numpy_array.dtype.itemsize
      return tuple(stride / itemsize for stride in strides_bytes)
      
    def expr_Strides():
      value = eval_expr(expr.array)
      return _strides(value)
    
    def expr_Attribute():
      value = eval_expr(expr.value)
      
      if expr.name == 'offset':
        if value.base is None:
          return 0
        else:
          offset_bytes =  value.ctypes.data - value.base.ctypes.data
          return offset_bytes / value.dtype.itemsize 
        
      elif expr.name == 'data':
        return np.ravel(value)
      
      elif expr.name == 'strides':
        return _strides(value)
      
      elif expr.name == 'step':
        step = getattr(value, 'step', 1)
        if step is None: step = 1 
        return step 
      
      elif isinstance(value, tuple):
        if expr.name.startswith('elt'):
          field = int(expr.name[3:])
        else:
          field = int(expr.name)
        return value[field]
        
      else:
        assert hasattr(value, expr.name), "Missing attribute %s from value %s" % (expr.name, value)
        return getattr(value, expr.name)

    def expr_Alloc():
      count = eval_expr(expr.count)
      arr = np.empty(shape = (count,), dtype = expr.elt_type.dtype)
      return arr
      
    def expr_AllocArray():

      shape = eval_expr(expr.shape)

      assert isinstance(shape, tuple), "Expected tuple, got %s" % (shape,)
      assert isinstance(expr.elt_type, ScalarT), \
          "Expected scalar element type for AllocArray, got %s" % (expr.elt_type,)
      dtype = expr.elt_type.dtype
      return  np.ndarray(shape = shape, dtype = dtype) 
    
    
    def expr_ArrayView():

      data = eval_expr(expr.data)
      
      shape  = eval_expr(expr.shape)
      strides = eval_expr(expr.strides)
      offset = eval_expr(expr.offset)
      dtype = expr.type.elt_type.dtype
      bytes_per_elt = dtype.itemsize
      byte_offset = offset * bytes_per_elt
      byte_strides = tuple(si * bytes_per_elt for si in  strides)
      if False:
        print expr 
        print "data", data
        print "shape(data)", np.shape(data)
      
        print "shape",  shape 
        print "strides", strides
        print "offset", offset 
        print "itemsize", bytes_per_elt
        print "byte strides", byte_strides 
        print "byte offset", byte_offset 
        
      if isinstance(data, np.ndarray):
        data = data.data 
 
      return np.ndarray(shape = shape, 
                        offset = byte_offset,  
                        buffer = data, 
                        strides = byte_strides, 
                        dtype = np.dtype(dtype))
      
      
    def expr_Array():
      elt_values = map(eval_expr, expr.elts)
      return np.array(elt_values)

    def expr_ConstArray():
      shape = eval_expr(expr.shape)
      value = eval_expr(expr.value)
      return np.ones(shape, dtype = expr.value.type.dtype) * value 
    
    def expr_ConstArrayLike():
      array = eval_expr(expr.array)
      value = eval_expr(expr.value)
      return np.ones_like(array, dtype = expr.value.type.elt_type.dtype) * value 
    
    def expr_TypeValue():
      t = expr.type_value  
      assert isinstance(t, ScalarT), \
        "Parakeet only supports scalar types as values, not %s" % expr 
      return t.dtype 
    
    def expr_Shape():
      return np.shape(eval_expr(expr.array))
    
    def expr_Reshape():
      array = eval_expr(expr.array)
      shape = eval_expr(expr.shape)
      return array.reshape(shape)
    
    def expr_Index():
      array = eval_expr(expr.value)
      index = eval_expr(expr.index)
      return array[index]

    def expr_PrimCall():
      arg_values = eval_args(expr.args)
      result = expr.prim.fn (*arg_values)
      return result 
    
    def expr_Slice():
      return slice(eval_expr(expr.start), eval_expr(expr.stop),
                   eval_expr(expr.step))

    def expr_Var():
      return env[expr.name]

    def expr_Call():
      fn = eval_expr(expr.fn)
      arg_values = eval_args(expr.args)
      return eval_fn(fn, arg_values)

    def expr_Closure():
      if isinstance(expr.fn, (UntypedFn, TypedFn)):
        fundef = expr.fn
      else:
        assert isinstance(expr.fn, str)
        fundef = UntypedFn.registry[expr.fn]
      closure_arg_vals = map(eval_expr, expr.args)
      return ClosureVal(fundef, closure_arg_vals)

    def expr_Fn():
      return ClosureVal(expr, [])

    def expr_TypedFn():
      return ClosureVal(expr, [])

    def expr_Cast():
      x = eval_expr(expr.value)
      t = expr.type
      assert isinstance(t, ScalarT)
      # use numpy's conversion function
      return t.dtype.type(x)
    
    def expr_Select():
      cond = eval_expr(expr.cond)
      trueval = eval_expr(expr.true_value)
      falseval = eval_expr(expr.false_value)
      return trueval if cond else falseval 

    def expr_Struct():
      assert expr.type, "Expected type on %s!" % expr
      assert isinstance(expr.type, StructT), \
          "Expected %s : %s to be a struct" % (expr, expr.type)
      elts = map(eval_expr, expr.args)
      return expr.type.ctypes_repr(*elts)

    def expr_Tuple():
      return tuple(map(eval_expr, expr.elts))
    
    def expr_TupleProj():
      return eval_expr(expr.tuple)[expr.index]

    def expr_ClosureElt():
      assert isinstance(expr.closure, Expr), \
          "Invalid closure expression-- %s : %s" % \
          (expr.closure, type(expr.closure))
      clos = eval_expr(expr.closure)
      return clos.fixed_args[expr.index]

    def expr_Range():
      return np.arange(eval_expr(expr.start), eval_expr(expr.stop), eval_expr(expr.step))
    
    def expr_Len():
      return len(eval_expr(expr.value))
    
    def normalize_axes(axis, args):
      if isinstance(axis, Expr):
        axis = eval_expr(axis)
        
      if isinstance(axis, tuple):
        axes = axis 
      else:
        axes = (axis,) * len(args)
      
      assert len(axes) == len(args)
    
    def expr_Map():
      fn = eval_expr(expr.fn)
      args = [eval_expr(arg) for arg in expr.args]
      axes = normalize_axes(expr.axis, args)
      
      largest_rank = 0
      largest_arg = None
      iter_space = None
      for arg, axis  in zip(args, axes):
        rank = rankof(arg) 
        if rank > largest_rank and (axis is None or axis < rank):
          largest_rank = rank 
          largest_arg = arg
          if axis is None: 
            iter_space = largest_arg.shape 
          else:
            iter_space = [largest_arg.shape[axis]]
          
      if largest_rank == 0:
        return eval_fn(fn, args)

      results = []
      for idx in np.ndindex(iter_space):
        elt_args = []
        for arg, axis in zip(args, axes):
          r = rankof(arg)
          if r == 0:
            elt_args.append(arg)
          if axis is None:
            curr_indices = idx[-(largest_rank - r):]
            elt_args.append(arg[tuple(curr_indices)])
          elif r > axis:
            indices = [slice(None) if j != axis else idx[0] for j in xrange(rankof(arg)) ]
            elt_args.append(arg[tuple(indices)])
          else:
            elt_args.append(arg)
        results.append(eval_fn(fn, elt_args))
      return np.array(results)
    
    def expr_Reduce():
      fn = eval_expr(expr.fn)
      combine = eval_expr(expr.combine)
      init = eval_expr(expr.init) if expr.init else None 
      args = [eval_expr(arg) for arg in expr.args]
      if isinstance(expr.axis, (int, long)):
        axis = expr.axis
      else:
        axis = eval_expr(expr.axis)
        
      if axis is None:
        args = [np.ravel(arg) for arg in args]
        axis = 0 
      
      largest_rank = 0
      largest_arg = None
      for arg in args:
        rank = rankof(arg) 
        if rank > largest_rank:
          largest_rank = rank 
          largest_arg = arg
          
      if largest_rank == 0:
        acc = eval_fn(fn, args)
        if init is None: return acc 
        else: return eval_fn(combine, [init, acc])
      
      niters = largest_arg.shape[axis]

      if init is not None: acc = init 
      else: acc = None 
        
      for i in xrange(niters):
        elt_args = []
        for arg in args:
          r = rankof(arg)
          if r > 0:
            indices = [slice(None) if j != axis else i for j in xrange(rankof(arg)) ]
            elt_args.append(arg[tuple(indices)])
          else:
            elt_args.append(arg)
        elt_result = eval_fn(fn, elt_args)
        if acc is not None:
          acc = eval_fn(combine, [acc, elt_result])
        else:
          acc = elt_result
      return acc
    
    def expr_Scan():
      assert False, "Scan not implemented"
    
    def expr_IndexMap():
      fn = eval_expr(expr.fn)
      shape = eval_expr(expr.shape)
      dtype = expr.type.elt_type.dtype
      result = np.empty(shape, dtype = dtype)
      for idx in np.ndindex(shape):
        result[idx] = eval_fn(fn, (idx,))
      return result
      
    def expr_IndexReduce():
      fn = eval_expr(expr.fn)
      combine = eval_expr(expr.combine)
      shape = eval_expr(expr.shape)
      if not isinstance(shape, (list, tuple) ):
        shape = [shape]
      ranges = [xrange(n) for n in shape]
      
      acc = eval_if_expr(expr.init)
      for idx in itertools.product(*ranges):
        if len(idx) == 1:
          idx = idx[0]
        elt = eval_fn(fn, (idx,))
        if acc is None:
          acc = elt 
        else:
          elt = eval_fn(combine, (acc, elt))
      return elt 
    
    fn_name = "expr_" + expr.__class__.__name__
    dispatch_fn = locals()[fn_name]
    result = dispatch_fn()
    
    # we don't support python function's inside parakeet,
    # they have to be translated into Parakeet functions
    if isinstance(result, types.FunctionType):
      fundef = ast_conversion.translate_function_value(result)
      return ClosureVal(fundef, fundef.python_nonlocals())
    else:
      return result

  def eval_merge_left(phi_nodes):
    for result, (left, _) in phi_nodes.iteritems():
      env[result] = eval_expr(left)

  def eval_merge_right(phi_nodes):
    for result, (_, right) in phi_nodes.iteritems():
      env[result] = eval_expr(right)

  def assign(lhs, rhs, env):
    if isinstance(lhs, Var):
      env[lhs.name] = rhs
    elif isinstance(lhs, Tuple):
      assert isinstance(rhs, tuple)
      for (elt, v) in zip(lhs.elts, rhs):
        assign(elt, v, env)
    elif isinstance(lhs, Index):
      arr = eval_expr(lhs.value)
      idx = eval_expr(lhs.index)

      arr[idx] = rhs

  

  def eval_parfor_seq(fn, bounds):
    if isinstance(bounds, (list,tuple)) and len(bounds) == 1:
      bounds = bounds[0]
        
    if isinstance(bounds, (int, long)):
      for idx in xrange(bounds):
        eval_fn(fn, (idx,))
    else:
      for idx in np.ndindex(bounds):
        eval_fn(fn, (idx,))
  
    
    
  def eval_stmt(stmt):

    if isinstance(stmt, Return):
      v = eval_expr(stmt.value)
      raise ReturnValue(v)
    
    elif isinstance(stmt, Assign):
      value = eval_expr(stmt.rhs)
      assign(stmt.lhs, value, env)

    elif isinstance(stmt, If):
      cond_val = eval_expr(stmt.cond)

      if cond_val:
        eval_block(stmt.true)
        
        eval_merge_left(stmt.merge)
      else:
        eval_block(stmt.false)
        eval_merge_right(stmt.merge)
        

    elif isinstance(stmt, While):
      eval_merge_left(stmt.merge)
      while eval_expr(stmt.cond):
        eval_block(stmt.body)
        eval_merge_right(stmt.merge)
        
    elif isinstance(stmt, ForLoop):
      start = eval_expr(stmt.start)
      stop = eval_expr(stmt.stop)
      step = eval_expr(stmt.step)
       
      eval_merge_left(stmt.merge)

      for i in xrange(start, stop, step):
        env[stmt.var.name] = i
        eval_block(stmt.body)
        eval_merge_right(stmt.merge)
      
        
    elif isinstance(stmt, ExprStmt):
      eval_expr(stmt.value)
      
    elif isinstance(stmt, ParFor):
      fn = eval_expr(stmt.fn)
      bounds = eval_expr(stmt.bounds)
    
      eval_parfor_seq(fn, bounds)    
      
      
    else:
      raise RuntimeError("Statement not implemented: %s" % stmt)

  def eval_block(stmts):

    for stmt in stmts:
      eval_stmt(stmt)

  try:
   
    eval_block(fn.body)

  except ReturnValue as r:
    return r.value 
  except:
    raise

def run_python_fn(python_fn, args, kwds):
  untyped  = ast_conversion.translate_function_value(python_fn)
  # should eventually roll this up into something cleaner, since
  # top-level functions are really acting like closures over their
  # global dependencies
  global_args = [python_fn.func_globals[n] for n in untyped.nonlocals]
  all_positional = global_args + list(args)
  actuals = args.FormalArgs(all_positional, kwds)
  return eval_fn(untyped, actuals)



import parakeet
import numpy as np 
from timer import compare_perf




def clamp(i, offset, maxval):
    j = max(0, i + offset)
    return min(j, maxval)


def reflect(pos, offset, bound):
    idx = pos+offset
    return min(2*(bound-1)-idx,max(idx,-idx))
 

def conv(x, weights, mode=clamp):
    sx = x.shape
    sw = weights.shape
    result = np.zeros_like(x)
    for i in xrange(sx[0]):
        for j in xrange(sx[1]):
            for ii in xrange(sw[0]):
                for jj in xrange(sw[1]):
                    idx = mode(i,ii-sw[0]/2,sw[0]), mode(j,jj-sw[0]/2,sw[0])
                    result[i,j] += x[idx] * weights[ii,jj] 
    return result


xsize = (300,300)
x = np.random.randn(*xsize)
wsize = (5,5)
w = np.random.randn(*wsize)

compare_perf(conv, [x,w])



import numpy as np 
from parakeet import testing_helpers 

def test_simple_linspace():
  testing_helpers.expect(np.linspace, [0.2, 0.9], np.linspace(0.2, 0.9))

def test_linspace_with_count():
  testing_helpers.expect(np.linspace, [-0.2, 0.9, 30], np.linspace(-0.2, 0.9, 30))

if __name__ == "__main__":
  testing_helpers.run_local_tests()





import numpy as np 
from parakeet import jit, testing_helpers


values = [1, 
          1.0,
          np.array([True, True, True]), 
          #np.array([1,2,3], dtype='int8'),
          np.array([1,2,3], dtype='int16'),
          #np.array([1,2,3], dtype='int32'), 
          #np.array([1,2,3], dtype='int64'),
          np.array([1,2,3], dtype='float32'), 

          #np.array([1,2,3], dtype='float64')
        ]

def run(parakeet_fn, python_fn):
  testing_helpers.expect_allpairs(jit(parakeet_fn), python_fn, values)
  
def add(x,y):
  return x + y

def test_add():
  run(add, np.add)

def sub(x,y):
  return x - y 

def test_sub():
  run(sub, np.subtract)
  
def mult(x,y):
  return x * y 

def test_mult():
  run(mult, np.multiply)
  
def div(x,y):
  return x / y 

def test_div():
  run(div, np.divide)
 
def mod(x,y):
  return x % y 

def test_mod():
  run(mod, np.mod)
 
if __name__ == '__main__':
  testing_helpers.run_local_tests()





from .. import ndtypes 
from .. ndtypes import ArrayT, Int64, elt_type, empty_tuple_t, TupleT, TypeValueT 
from .. frontend import macro, jit, typed_macro 
from .. syntax import (Attribute, Tuple, Ravel, Shape, Reshape, TypeValue, Transpose, Const)
from .. syntax.helpers import const_int, zero_i64 


@macro
def transpose(x):
  return Transpose(x)

@macro 
def ravel(x):
  return Ravel(x)
  
@macro 
def reshape(x):
  return Reshape(x)


@typed_macro 
def get_elt_type(x):
  elt_t = ndtypes.elt_type(x.type)
  return TypeValue(elt_t, type = TypeValueT(elt_t))
  
@typed_macro
def itemsize(xt):
  return const_int(elt_type(xt.type).nbytes)
  
@typed_macro 
def rank(xt):
  return const_int(xt.type.rank) 

@typed_macro 
def size(xt, axis = None):
  if axis is None: 
    axis = zero_i64
  assert isinstance(axis, Const), "Axis argument to 'size' must be a constant, given %s" % axis
  assert axis.value == 0, "Calling 'size' along axes other than 0 not yet supported, given %s" % axis
  if isinstance(xt.type, ArrayT):
    return Attribute(xt, 'size', type = Int64)
  elif isinstance(xt.type, TupleT):
    return const_int(len(xt.type.elt_types))
  else:
    return const_int(1)


@macro  
def shape(x):
  return Shape(x)




from .. import prims
from ..ndtypes import FloatT, Float32, Float64 


# names used in math library 
_float_fn_names = {
  prims.tan : 'tan', 
  prims.tanh : 'tanh',
  prims.arctan : 'atan', 
  prims.arctanh : 'atanh',
  prims.arctan2 : 'atan2',  
   
  prims.cos : 'cos', 
  prims.cosh : 'cosh', 
  prims.arccos : 'acos', 
  prims.arccosh : 'acosh', 
  
  prims.sin : 'sin', 
  prims.sinh : 'sinh',
  prims.arcsin : 'asin',
  prims.arcsinh : 'asinh', 
  
  prims.log : 'log',
  prims.log2 : 'log2', 
  prims.log10 : 'log10',
  prims.log1p : 'log1p', 
   
   
  prims.exp : 'exp',  
  prims.exp2 : 'exp2', 
  prims.power : 'pow', 
  prims.expm1 : 'expm1', 
  prims.sqrt : 'sqrt', 
  
  prims.abs : 'fabs', 
  
  prims.ceil : 'ceil', 
  prims.floor : 'floor', 
  prims.round : 'round', 
  
} 
 
def float_prim(p, t):
  assert p in _float_fn_names
  assert isinstance(t, FloatT)
  name = _float_fn_names[p]
  if t == Float32:
    return name + "f" 
  else:
    return name 
  



import numpy as np 
from parakeet.testing_helpers import expect_each, run_local_tests

size = (5,5)
float_mat = np.random.uniform(0,1,size=size)
int_mat = np.random.random_integers(0,255,size=size)

matrices = [float_mat, int_mat]

def diff_x(I):
  m = I.shape[0]
  return (I[1:, :] - I[:m-1, :])[:, 1:]

def test_diff_x():
  expect_each(diff_x, diff_x, matrices)

def diff_y(I):
  n = I.shape[1]
  return (I[:, 1:] - I[:, :n-1])[1:, :]

def test_diff_y():
  expect_each(diff_x, diff_x, matrices)

if __name__ == "__main__":
  run_local_tests()




import numpy as np

from parakeet import jit, config 
from parakeet.testing_helpers import expect, run_local_tests 



@jit 
def arange1(a):
  return np.arange(a)

@jit  
def arange2(a,b):
  return np.arange(a,b)

@jit  
def arange3(a,b,c):
  return np.arange(a,b,c)

def test_range1_int():
  a = 5
  expect(arange1, [a], np.arange(a))

def test_range1_float():
  a = 3.5
  expect(arange1, [a], np.arange(a))
  
def test_range2_int():
  a = 2
  b = 6
  expect(arange2, [a,b], np.arange(a,b))

def test_range2_float():
  a = 3.5
  b = 7.1
  expect(arange2, [a,b], np.arange(a,b))
  
def test_range3_int():
  a = 2
  b = 10
  c = 3
  expect(arange3, [a,b,c], np.arange(a,b,c))

def test_range3_float():
  a = 3.5
  b = 9.1
  c = 1.7
  expect(arange3, [a,b,c], np.arange(a,b,c))

def test_range3_int_reverse():
  a = 10
  b = 2
  c = -3
  expect(arange3, [a,b,c], np.arange(a,b,c))

def test_range3_float_reverse():
  a = 9.1
  b = 3.5
  c = -1.7
  expect(arange3, [a,b,c], np.arange(a,b,c))  
    


if __name__ == '__main__':
  run_local_tests()




from ..analysis.value_range_analysis import TupleOfIntervals, SliceOfIntervals
from ..ndtypes import ArrayT, TupleT, SliceT, ScalarT, NoneType 
from ..syntax import Tuple, Slice, Var, Index  
from ..syntax.helpers import zero_i64, one_i64, slice_none, make_slice_type, none

from range_transform import RangeTransform, Interval, NoneValue

class NegativeIndexElim(RangeTransform):
  
  """
  This doesn't entirely obviate the need for runtime checks on indexing 
  but does handle the simple case where a negative index can be inferred from the source 
  """
  
  def has_negative(self, range_value, inclusive = False):
    if isinstance(range_value, Interval):
      return self.always_negative(range_value, inclusive = inclusive)
    elif isinstance(range_value, TupleOfIntervals):
      return any(self.has_negative(elt) for elt in range_value.elts)
    elif isinstance(range_value, SliceOfIntervals):
      return self.has_negative(range_value.start) or \
        self.has_negative(range_value.stop) or \
        self.has_negative(range_value.step) 
    
  
  def transform_Index(self, expr):

    arr = expr.value 
    arr_t = arr.type 
    if arr_t.__class__ is not ArrayT:
      return expr 
    
    index = expr.index 
    
    if isinstance(index.type, (SliceT, ScalarT)):
      assert arr_t.rank > 0, "Unexpected zero-rank array in %s" % expr 
      extra_indices = [slice_none] * (arr_t.rank - 1) 
      index_elts = [index] + extra_indices
    elif isinstance(index.type, TupleT):
      n_given = len(index.type.elt_types)
      n_expected = arr_t.rank  
      assert n_given <= n_expected, "Too many indices in %s" % expr 
      
      # replace None in the indexing with Slice(None,None,None)
      given_elts = [given if given.type != NoneType else slice_none 
                    for given in self.tuple_elts(index)]
      index_elts = given_elts + [slice_none] * (n_expected - n_given)
      
    index_elts = list(index_elts)
    index_ranges = [self.get(idx_elt) for idx_elt in index_elts]
    
    assert len(index_ranges) == len(index_elts)
    shape = self.shape(expr.value, temp = False)
    shape_elts = self.tuple_elts(shape)
  
    assert len(shape_elts) == len(index_elts), \
      "Mismatch between number of indices %d and rank %d in %s"  % \
      (len(index_elts), len(shape_elts), expr)
    for i, (index_range, index_elt) in enumerate(zip(index_ranges, index_elts)):
      if isinstance(index_range, Interval):
        if index_range.upper == index_range.lower:
          index_elt = self.int(index_range.upper, "const_idx")
          
        if self.always_negative(index_range, inclusive=False):
          index_elts[i] = self.add(shape_elts[i], index_elt)
         
      elif isinstance(index_range, SliceOfIntervals) and self.has_negative(index_range):
        
        if isinstance(index_range.start, NoneValue):
          start = none
        elif index_range.start.upper == index_range.start.lower:
          start = self.int(index_range.start.lower)
        else:
          start = self.attr(index_elt, 'start', temp = False)
        
        if isinstance(index_range.stop, NoneValue):
          stop = none 
        elif index_range.stop.upper == index_range.stop.lower:
          stop = self.int(index_range.stop.lower)
        else:
          stop = self.attr(index_elt, 'stop', temp = False)
        
        if isinstance(index_range.step, NoneValue):
          step = none 
        elif index_range.step.lower == index_range.step.upper:
          step = self.int(index_range.step.lower)
        else: 
          step = self.attr(index_elt, 'step', temp = False)
          
        shape_elt = shape_elts[i]
        
        if self.always_negative(index_range.start, inclusive=False):
          start = self.add(shape_elt, start, "start")
        
        if self.always_negative(index_range.stop, inclusive=False):
          stop = self.add(shape_elt, stop, "stop")
          
        if self.always_negative(index_range.step):
          if self.is_none(start):
            start = self.sub(shape_elt, one_i64, "last_pos")
          
          #if self.is_none(stop):
          # if it's a negative step then (stupidly) there's
          # no valid stop value
          if self.is_none(stop):
            # from this point forward, negative stop just means include the element 0
            stop = self.int(-1)
            #assert False, \
            #  "Not yet supported: Python's peculiar behavior with negative slice steps and None stop"
          
            
        else:
          if self.is_none(step):
            step = one_i64 
            
          if self.is_none(start):
            start = zero_i64
            
          if self.is_none(stop):
            stop = shape_elt
        
        
        slice_t = make_slice_type(start.type, stop.type, step.type)
        index_elts[i] = Slice(start=start, stop=stop, step = step, type = slice_t)
      if len(index_elts) > 1:
        index = self.tuple(index_elts)
      else:
        index = index_elts[0]
      expr.index = index  

    return expr 
          
    
    
    
        


from itertools import izip 
import numpy as np
from ..ndtypes import (type_conv, ScalarT, ArrayT, FnT, ClosureT, SliceT, NoneT, TupleT, TypeValueT)
from ..syntax import TypedFn, UntypedFn


def prepare_closure_args(untyped_fn):
  closure_args = untyped_fn.python_nonlocals()
  closure_arg_types = [type_conv.typeof(v) for v in closure_args]
  return prepare_args(closure_args, closure_arg_types)
      

def prepare_arg(arg, t):
  if isinstance(t, ScalarT):
    return t.dtype.type(arg)
  elif isinstance(t, ArrayT):
    return np.asarray(arg)
  elif isinstance(t, TupleT):
    arg = tuple(arg)
    assert len(arg) == len(t.elt_types)
    return prepare_args(arg, t.elt_types)
  elif isinstance(t, (NoneT, SliceT)):
    return arg
  elif isinstance(t, TypeValueT):
    return ()
  elif isinstance(t, (FnT, ClosureT)):
    if isinstance(arg, TypedFn):
      return ()
    elif isinstance(arg, UntypedFn):
      return prepare_closure_args(arg)
    else:
      from ..frontend import ast_conversion
      untyped = ast_conversion.translate_function_value(arg)
      return prepare_closure_args(untyped)
  assert False, "Can't call compiled C code with argument %s (Python type = %s, Parakeet type = %s)" % \
    (arg, type(arg), t)
  
def prepare_args(args, arg_types):
  return tuple(prepare_arg(arg, t) for arg, t in izip(args, arg_types))
  




import numpy as np
import parakeet
from parakeet import shape_inference
from parakeet.shape_inference import call_shape_expr
from parakeet.shape_inference.shape import const, Shape, Var 
from parakeet import testing_helpers

def expect_shape(python_fn, args_list, expected):
  print "[expect_shape]"
  print " -- fn: ", python_fn
  print " -- args: ", args_list
  typed_fn = parakeet.typed_repr(python_fn, args_list)
  print " -- types: ", typed_fn.input_types
  result_shape = call_shape_expr(typed_fn)
  assert result_shape == expected, \
      "Expected shape %s, but got: %s" % (expected, result_shape)

def const_scalar():
  return 1

def test_const_scalar():
  expect_shape(const_scalar, [], const(1))

def merge_scalars(b):
  if b:
    return 1
  else:
    return 2

def test_any_scalar():
  expect_shape(merge_scalars, [True], shape_inference.any_scalar)

def array_literal():
  return [1,2,3]

def test_array_literal():
  expect_shape(array_literal, [], Shape([3]))

vec = np.array([1,2,3])
mat = np.array([[1,2,3],[4,5,6],[7,8,9]])

def ident(x):
  return x

def test_ident_1d():
  expect_shape(ident, [vec], Shape([Var(0)]))

def test_ident_2d():
  expect_shape(ident, [mat], Shape([Var(0), Var(1)]))

def increase_rank(x):
  return [x,x]

def test_increase_rank_1d():
  expect_shape(increase_rank, [1], Shape([2]))

def test_increase_rank_2d():
  expect_shape(increase_rank, [vec], Shape([2, Var(0)]))

def test_increase_rank_3d():
  expect_shape(increase_rank, [mat], Shape([2, Var(0), Var(1)]))

def incr(xi):
  return xi + 1


def simple_map(x):
  return parakeet.map(incr, x)

def test_simple_map_1d():
  expect_shape(simple_map, [vec], Shape([Var(0)]))

def test_simple_map_2d():
  expect_shape(simple_map, [mat], Shape([Var(0), Var(1)]))

def map_increase_rank(x):
  return parakeet.map(increase_rank, x)

def test_map_increase_rank_1d():
  expect_shape(map_increase_rank, [vec], Shape([Var(0), 2]))

def test_map_increase_rank_2d():
  expect_shape(simple_map, [mat], Shape([Var(0), Var(1)]))

if __name__ == '__main__':
  testing_helpers.run_local_tests()



import numpy as np 
import parakeet 

def wrap(pos, offset, bound):
    return ( pos + offset ) % bound

def clamp(pos, offset, bound):
    return min(bound-1,max(0,pos+offset))

def reflect(pos, offset, bound):
    idx = pos+offset
    return min(2*(bound-1)-idx,max(idx,-idx))


def python_local_maxima(data, wsize, mode=wrap):
  result = np.ones(shape=data.shape,dtype='bool')
  for pos in np.ndindex(data.shape):
    myval = data[pos]  
    for offset in np.ndindex(wsize):
      neighbor_idx = tuple(mode(p, o-w/2, w) for (p, o, w) in zip(pos, offset, wsize))
      result[pos] &= (data[neighbor_idx] <= myval)
  return result 


@parakeet.jit 
def parakeet_local_maxima(data, wsize, mode=wrap):
  def is_max(pos):
    def is_smaller_neighbor(offset):
      neighbor_idx = tuple(mode(p, o-w/2, w) for (p, o, w) in zip(pos, offset, wsize))
      return data[neighbor_idx] <= data[pos]
    return np.all(parakeet.imap(is_smaller_neighbor, wsize))
  return parakeet.imap(is_max, data.shape)
  

# not sure how to get numba to auto-jit size generic code
# get error: "FAILED with KeyError 'sized_pointer(npy_intp, 4)'"
#import numba
#numba_local_maxima = numba.autojit(python_local_maxima) 

if __name__  == '__main__':
  from timer import timer 
  shape = (30,30,20,12)
  x = np.random.randn(*shape)
  wsize = (3,3,3,3)
  with timer("Parakeet C (first run)"):
    parakeet_result_c  = parakeet_local_maxima(x, wsize, _backend = 'c')
  with timer("Parakeet C (second run)"):
    parakeet_result_c  = parakeet_local_maxima(x, wsize, _backend = 'c')
  
  with timer("Parakeet OpenMP (first run)"):
    parakeet_result_omp  = parakeet_local_maxima(x, wsize, _backend = 'openmp')
  assert np.allclose(parakeet_result_c, parakeet_result_omp), \
    "OpenMP backend differs from C: %s vs. %s" % (np.sum(parakeet_result_c), np.sum(parakeet_result_omp))
  with timer("Parakeet OpenMP (second run)"):
    parakeet_result_omp  = parakeet_local_maxima(x, wsize, _backend = 'openmp')

  with timer("Parakeet CUDA (first run)"):
    parakeet_result_cuda  = parakeet_local_maxima(x, wsize, _backend = 'cuda')
  assert np.allclose(parakeet_result_c, parakeet_result_cuda), \
    "CUDA backend differs from C: %s vs %s" % (np.sum(parakeet_result_c), np.sum(parakeet_result_cuda))
  with timer("Parakeet CUDA (second run)"):
    parakeet_result_cuda  = parakeet_local_maxima(x, wsize, _backend = 'cuda')

  #with timer("Numba (first run)"):
  #  numba_result  = numba_local_maxima(x, wsize, wrap)
  #with timer("Numba (second run)"):
  #  numba_result  = numba_local_maxima(x, wsize, wrap)
  with timer("Python"):
    python_result = python_local_maxima(x, wsize) 
  
  assert np.allclose(python_result,parakeet_result)
  
  



from prepare_args import prepare_args
from ..transforms.pipeline  import loopify, final_loop_optimizations  
from ..value_specialization import specialize
from ..config import value_specialization
from pymodule_compiler import PyModuleCompiler 



_cache = {}
def run(fn, args):
  args = prepare_args(args, fn.input_types)
  
  transformed_fn = loopify.apply(fn)
  transformed_fn = final_loop_optimizations.apply(transformed_fn)
  if value_specialization: 
    transformed_fn = specialize(transformed_fn, args)

  key = transformed_fn.cache_key
  if key in _cache:
    return _cache[key](*args)
  compiled_fn = PyModuleCompiler().compile_entry(transformed_fn)
  c_fn = compiled_fn.c_fn 
  _cache[key] = c_fn 
  return c_fn(*args)
  


from ..builder import build_fn
from ..ndtypes import ArrayT, lower_rank
from ..prims import Prim  
from ..syntax import Return, Map 
from ..syntax.helpers import is_identity_fn, unwrap_constant
from transform import Transform

def get_nested_map(fn):
  if len(fn.body) != 1:
    return None 
  stmt = fn.body[0]
  if stmt.__class__ is not Return:
    return None 
  if stmt.value.__class__ is not Map:
    return None 
  return stmt.value 
     

class PermuteReductions(Transform):
  """
  When we have a reduction of array values, such as:
     Reduce(combine = Map(f), X, axis = 0) 
  it can be more efficient to interchange the Map and Reduce:
     Map(combine = f, X, axis = 1)
  """
  
  
  def transform_Reduce(self, expr):
    return expr 
  
  def transform_Scan(self, expr):
    if len(expr.args) > 1:
      return expr 
    
    axis = unwrap_constant(expr.axis)
    if axis is None or not isinstance(axis, (int,long)) or axis > 1 or axis < 0:
      return expr
     
    if not isinstance(expr.type, ArrayT) or expr.type.rank != 2:
      return expr
    
    fn = self.get_fn(expr.fn)
    fn_closure_args = self.closure_elts(expr.fn)
    if len(fn_closure_args) > 0:
      return expr 
    
    combine = self.get_fn(expr.combine)
    combine_closure_args = self.closure_elts(expr.closure)
    if len(combine_closure_args) > 0:
      return expr 
    
    if is_identity_fn(fn):
      nested_map = get_nested_map(combine)
    
    if not isinstance(nested_map.fn, Prim):
      return expr  
    
    arg_t = expr.args[0].type
    elt_t = lower_rank(arg_t, 1)
    new_nested_fn = None 
    return Map(fn = new_nested_fn,  
               args = expr.args, 
               axis = 1 - axis, 
               type = expr.type)
      
       
  
    



import numpy as np 

from .. import prims 
from ..frontend import typed_macro 
from ..ndtypes import ScalarT, ArrayT, make_array_type
from ..syntax import PrimCall, Call,  OuterMap, Map
from ..syntax.helpers import make_closure 

def _get_vdot_fn(a, b):
  from reductions  import vdot  
  from ..frontend import ast_conversion 
  vdot_untyped = ast_conversion.translate_function_value(vdot)
  from ..type_inference import specialize
  vec_type_a = make_array_type(a.type.elt_type, 1) 
  vec_type_b = make_array_type(b.type.elt_type, 1)
  result_scalar_type = a.type.elt_type.combine(b.type.elt_type)
 
  vdot_typed = specialize(vdot_untyped, [vec_type_a, vec_type_b] )

  assert vdot_typed.return_type == result_scalar_type, \
    "Expected return type %s but got %s from vdot" % (result_scalar_type, vdot_typed.return_type)
  return vdot_typed 

@typed_macro
def dot(a,b):
  if isinstance(a.type, ScalarT):
    return PrimCall(prims.multiply, [a, b], type = b.type.combine(a.type))
  elif isinstance(b.type, ScalarT):
    return PrimCall(prims.multiply, [a, b], type = a.type.combine(b.type))
  
  assert isinstance(a.type, ArrayT), "Expected %s to be array but got %s" % (a, a.type)
  assert isinstance(a.type, ArrayT), "Expected %s to be array but got %s" % (b, b.type)
      
      
  if a.type.rank == 1 and b.type.rank == 1:
    vdot = _get_vdot_fn(a,b)
    return Call(fn = vdot, args = [a, b], type = vdot.return_type)

  elif a.type.rank == 1:
    vdot = _get_vdot_fn(a,b)
    
    assert b.type.rank == 2, "Don't know how to multiply %s and %s" % (a.type, b.type)
    vdot_col = make_closure(vdot, (a,))
    result_vec_type = make_array_type(vdot.return_type, 1)
    return Map(fn = vdot_col, args = (b,), axis = 1, type = result_vec_type)
        
  elif b.type.rank == 1:
    assert a.type.rank == 2, "Don't know how to multiply %s and %s" % (a.type, b.type)
    vdot = _get_vdot_fn(b,a)
    vdot_row = make_closure(vdot, (b,))
    result_vec_type = make_array_type(vdot.return_type, 1)
    return Map(fn = vdot_row, args = (a,), axis = 0, type = result_vec_type)
  else:  
    assert a.type.rank == 2 and b.type.rank == 2, \
        "Don't know how to multiply %s and %s" % (a.type, b.type)
    vdot = _get_vdot_fn(a,b)
    result_matrix_type = make_array_type(vdot.return_type, 2)
    return OuterMap(fn = vdot, args = (a, b), axis = (0,1), 
                    type = result_matrix_type)
      
@typed_macro 
def norm(x, ord=None):
  if ord is None:
    ord = 2
  if isinstance(x.type, ScalarT):
    return x
  assert isinstance(x.type, ArrayT), "Argument to 'norm' must be array, got %s : %s" % (x, x.type)
  assert x.type.rank == 1, "Norm currently only supported for vectors, not for %s : %s" (x, x.type)
  if ord == 0:
    return (x != 0).sum()
  elif ord == 1:
    return abs(x).sum()
  elif ord == 2:
    # not using 'conj' so won't yet work for complex numbers 
    return np.sqrt((x**2).sum())
  else:
    return (abs(x)**ord).sum() ** (1.0 / ord)

  



import __builtin__
import numpy as np 

from .. frontend.decorators import jit, macro, typed_macro 
from .. ndtypes import ( make_array_type,  combine_type_list, repeat_tuple,  
                         ArrayT, TupleT,  Int64, ScalarT) 
from .. syntax import (Range, Cast, AllocArray, TupleProj, Array, 
                       ConstArray, ConstArrayLike, Index, Shape)

from ..syntax.helpers import get_types, one_i64, zero_i64, make_tuple

from adverbs import imap 
from numpy_types import float64
from lib_helpers import _get_type, _get_shape, _get_tuple_elts


@macro
def arange(n, *xs, **kwds):
  if 'dtype' in kwds:
    elt_t = _get_type(kwds['dtype'])
    del kwds['dtype']
  else:
    elt_t = Int64
    
  assert len(kwds) == 0, "Unexpected keyword arguments to 'arange': %s" % kwds
  array_t = make_array_type(elt_t, 1) 
  count = __builtin__.len(xs)
  assert 0 <= count <= 2, "Too many args for range: %s" % ((n,) + tuple(xs))
 
  if count == 0:
    start = zero_i64 
    stop = n 
    step = one_i64 
  elif count == 1:
    start = n 
    stop = xs[0]
    step = one_i64 
  else:
    start = n 
    stop = xs[0]
    step = xs[1]
    
  if elt_t != Int64:
    start = Cast(start, type = elt_t)
    stop = Cast(stop, type = elt_t)
    step = Cast(step, type = elt_t)
  return Range(start, stop, step, type = array_t)
 
@typed_macro
def empty(shape, dtype = float64):
  elt_t = _get_type(dtype)
  assert isinstance(elt_t, ScalarT), "Array element type %s must be scalar" % (elt_t,)
  shape = _get_shape(shape)  
  rank = len(shape.type.elt_types)
  arr_t = make_array_type(elt_t, rank)
  return AllocArray(shape = shape, elt_type = elt_t, type = arr_t)

@jit 
def empty_like(x, dtype = None):
  if dtype is None:
    return empty(x.shape, x.dtype)
  else:
    return empty(x.shape, dtype)
  
@typed_macro   
def zeros(shape, dtype = float64):
  shape = _get_shape(shape)
  elt_type = _get_type(dtype)
  zero = Cast(zero_i64, type = elt_type)
  ndims = len(shape.type.elt_types)
  if ndims == 0:
    return zero 
  else:
    t = make_array_type(elt_type, ndims)
    return ConstArray(shape = shape, value = zero, type = t)
  
@jit
def zeros_like(x, dtype = None):
  if dtype is None:
    dtype = x.dtype
  return zeros(x.shape, dtype)

@typed_macro
def ones(shape, dtype = float64):
  shape = _get_shape(shape)
  elt_type = _get_type(dtype)
  one = Cast(one_i64, type = elt_type)
  ndims = len(shape.type.elt_types)
  if ndims == 0:
    return one 
  else:
    t = make_array_type(elt_type, ndims)
    return ConstArray(shape = shape, value = one, type = t)

@jit  
def ones_like(x, dtype = None):

  if dtype is None:
    dtype = x.dtype
  return ones(x.shape, dtype)

@jit
def copy(x):
  return [xi for xi in x]

@jit 
def linspace(start, stop, num = 50, endpoint = True):
  """
  Copied from numpy.linspace but dropped the 'retstep' option 
  which allows you to optionall return a tuple (that messes with type inference)
  """
  num = int(num)
  if num <= 0:
    return np.array([])
  elif endpoint:
    if num == 1:
      return np.array([float(start)])
    step = (stop-start)/float((num-1))
    y = np.arange(0, num) * step + start
    y[-1] = stop
    return y
  else:
    step = (stop-start)/float(num)
    return np.arange(0, num) * step + start 



@typed_macro
def array(value, dtype = None):
  if dtype is not None:
    expected_elt_type = _get_type(dtype)
    print "Got dtype argument to array: %s, ignoring for now" % expected_elt_type

  if isinstance(value.type, ArrayT):
    return value 
  else:
    assert isinstance(value.type, TupleT), "Don't know how to make array from %s : %s" % (value, value.type)
    elt_types = value.type.elt_types
    assert all(isinstance(t, ScalarT) for t in elt_types), \
      "Can only make array from tuple of scalars, not %s : %s" % (value, value.type)
    elt_t = combine_type_list(value.type.elt_types)
    array_elts = _get_tuple_elts(value, cast_type = elt_t)
    array_t = make_array_type(elt_t, 1)
    return Array(elts = array_elts, type = array_t)

@typed_macro
def tile(A, reps):
  reps = _get_shape(reps)
  reps_dims = len(reps.type.elt_types)
  if reps_dims == 0:
    return A 
  A_rank = A.type.rank if isinstance(A.type, ArrayT) else 0 
  if A_rank == 0:
    # array scalars, ugh!
    if isinstance(A.type, ArrayT):
      A = Index(A, make_tuple(()), type = A.elt_type)
    assert isinstance(A.type, ScalarT), "First argument to 'tile' must be array or scalar"
    array_t = make_array_type(A.type, reps_dims)
    return ConstArray(value = A, shape = reps, type = array_t)
  else:
    A_shape = Shape(A, type = repeat_tuple(Int64, A_rank))
    A_shape_elts = _get_tuple_elts(A_shape)
    reps_elts = _get_tuple_elts(reps)
    result_shape_elts = []
    assert False, "np.tile not yet implemented"
    #for i in xrange(max(len(A_shape_elts), len(reps_elts))):
      
    #result = AllocArray()
    
  


from clone_stmt import CloneStmt
from clone_function import CloneFunction
from dead_code_elim import DCE  
from inline import Inliner
from phase import Phase 
from simplify import Simplify
from subst import subst_expr, subst_expr_list, subst_expr_tuple, subst_stmt_list 
from transform import Transform


import time

from .. import config
from .. analysis import verify
from .. builder import Builder  
from .. syntax import (Expr, If, Assign, While, Return, ExprStmt, ForLoop, Comment, ParFor, 
                       Var, Tuple, Index, Attribute, Const, PrimCall, Struct, Alloc, Cast,  
                       TupleProj, Slice, ArrayView, Call, TypedFn,  AllocArray, Len, UntypedFn,  
                       Map, Reduce) 

transform_timings = {}
transform_counts = {}
if config.print_transform_timings:
  import atexit
  def print_timings():
    print "TRANSFORM TIMINGS"
    items = transform_timings.items()
    items.sort(key = lambda (_,t): t)
    items.reverse()
    for k, t in items:
      count = transform_counts[k]
      print "  %30s  Total = %6dms, Count = %4d, Avg = %3fms" % \
            (k.__name__, t*1000, count, (t/count)*1000)
  atexit.register(print_timings)

class Transform(Builder):
  def __init__(self, verify=config.opt_verify,
                     reverse=False,
                     require_types=True):

    Builder.__init__(self)
    self.fn = None
    self.verify = verify
    self.reverse = reverse
    self.require_types = require_types

  def __str__(self):
    return self.__class__.__name__ 
  
  def __repr__(self):
    return str(self)
  
  def __hash__(self):
    return hash(str(self))
  
  def __eq__(self, other):
    return str(self) == str(other)
  
  def lookup_type(self, name):
    assert self.type_env is not None
    return self.type_env[name]

  def transform_if_expr(self, maybe_expr):
    if isinstance(maybe_expr, Expr):
      return self.transform_expr(maybe_expr)
    elif isinstance(maybe_expr, tuple):
      return tuple([self.transform_if_expr(x) for x in maybe_expr])
    elif isinstance(maybe_expr, list):
      return [self.transform_if_expr(x) for x in maybe_expr]
    #elif hasattr(maybe_expr, 'transform'):
    #  return maybe_expr.transform(self.transform_expr)
    else:
      return maybe_expr

  def transform_generic_expr(self, expr):
    for member_name in expr.members():
      old_value = getattr(expr, member_name)
      new_value = self.transform_if_expr(old_value)
      setattr(expr, member_name, new_value)
    return expr

  def find_method(self, expr, prefix = "transform_"):
    assert expr, "Expected expression but got %s" % expr 
    method_name = prefix + expr.__class__.__name__
    if hasattr(self, method_name):
      return getattr(self, method_name)
    else:
      return None

  """
  Common cases for expression transforms: we don't need to create a method for
  every sort of expression but these run faster and allocate less memory than
  transform_generic_expr
  """
  def transform_Var(self, expr):
    return expr

  def transform_Tuple(self, expr):
    expr.elts = tuple(self.transform_expr(elt) for elt in expr.elts)
    return expr

  def transform_Const(self, expr):
    return expr

  def transform_Index(self, expr):
    expr.value = self.transform_expr(expr.value)
    expr.index = self.transform_expr(expr.index)
    return expr

  def transform_Attribute(self, expr):
    expr.value = self.transform_expr(expr.value)
    return expr

  def transform_PrimCall(self, expr):
    expr.args = self.transform_expr_tuple(expr.args)
    return expr

  def transform_Call(self, expr):
    expr.fn = self.transform_expr(expr.fn)
    expr.args = self.transform_expr_tuple(expr.args)
    return expr

  def transform_Alloc(self, expr):
    expr.count = self.transform_expr(expr.count)
    return expr

  def transform_Struct(self, expr):
    expr.args = self.transform_expr_tuple(expr.args)
    return expr

  def transform_Cast(self, expr):
    expr.value = self.transform_expr(expr.value)
    return expr

  def transform_Select(self, expr):
    expr.cond = self.transform_expr(expr.cond)
    expr.true_value = self.transform_expr(expr.true_value)
    expr.false_value = self.transform_expr(expr.false_value)
    return expr 

  def transform_TupleProj(self, expr):
    expr.tuple = self.transform_expr(expr.tuple)
    return expr

  def transform_TypedFn(self, expr):
    """By default, don't do recursive transformation of referenced functions"""
    return expr
  
  def transform_UntypedFn(self, expr):
    """By default, don't do recursive transformation of referenced functions"""
    return expr


  def transform_Slice(self, expr):
    expr.start = self.transform_expr(expr.start) if expr.start else None
    expr.stop = self.transform_expr(expr.stop) if expr.stop else None
    expr.step = self.transform_expr(expr.step) if expr.step else None
    return expr
  
  def transform_Range(self, expr):
    expr.start = self.transform_expr(expr.start) if expr.start else None
    expr.stop = self.transform_expr(expr.stop) if expr.stop else None
    expr.step = self.transform_expr(expr.step) if expr.step else None
    return expr
    

  def transform_Array(self, expr):
    expr.elts = self.transform_expr_tuple(expr.elts) 
    return expr
    
  def transform_ArrayView(self, expr):
    expr.data = self.transform_expr(expr.data)
    expr.shape = self.transform_expr(expr.shape)
    expr.strides = self.transform_expr(expr.strides)
    expr.offset = self.transform_expr(expr.offset)
    expr.size = self.transform_expr(expr.size)
    return expr
  
  def transform_AllocArray(self, expr):
    expr.shape = self.transform_expr(expr.shape)
    return expr 
  
  def transform_ConstArray(self, expr):
    expr.shape = self.transform_expr(expr.shape)
    expr.value = self.transform_expr(expr.value)
    
  def transform_ConstArrayLike(self, expr):
    expr.array = self.transform_expr(expr.array)
    expr.value = self.transform_expr(expr.value)

  def transform_Ravel(self, expr):
    expr.array = self.transform_expr(expr.array)
    return expr 
  
  def transform_Reshape(self, expr):
    expr.array = self.transform_expr(expr.array)
    expr.shape = self.transform_expr(expr.shape)
    return expr
  
  def transform_Transpose(self, expr):
    expr.array = self.transform_expr(expr.array)
    return expr 
  
  def transform_Shape(self, expr):
    expr.array = self.transform_expr(expr.array)
    return expr 
  
  def transform_Len(self, expr):
    expr.value = self.transform_expr(expr.value) 
    return expr

  def transform_IndexMap(self, expr):
    expr.fn = self.transform_expr(expr.fn)
    expr.shape = self.transform_expr(expr.shape)
    return expr 
  
  def transform_IndexReduce(self, expr):
    expr.fn = self.transform_expr(expr.fn)
    expr.combine = self.transform_if_expr(expr.combine)
    expr.shape = self.transform_expr(expr.shape)
    expr.init = self.transform_if_expr(expr.init)
    return expr
  
  def transform_IndexScan(self, expr):
    expr.fn = self.transform_expr(expr.fn)
    expr.combine = self.transform_if_expr(expr.combine)
    expr.shape = self.transform_expr(expr.shape)
    expr.init = self.transform_if_expr(expr.init)
    return expr
  
  def transform_Map(self, expr):
    expr.axis = self.transform_if_expr(expr.axis)
    expr.fn = self.transform_expr(expr.fn)
    expr.args = self.transform_expr_list(expr.args)
    return expr
  
  def transform_Reduce(self, expr):
    expr.axis = self.transform_if_expr(expr.axis)
    expr.init = self.transform_if_expr(expr.init)
    expr.fn = self.transform_expr(expr.fn)
    expr.combine = self.transform_if_expr(expr.combine)
    expr.args = self.transform_expr_list(expr.args)
    return expr
  
  def transform_Scan(self, expr):
    expr.axis = self.transform_if_expr(expr.axis)
    expr.init = self.transform_if_expr(expr.init)
    expr.fn = self.transform_expr(expr.fn)
    expr.combine = self.transform_if_expr(expr.combine)
    expr.args = self.transform_expr_list(expr.args)
    expr.emit = self.transform_expr(expr.emit)
    return expr
  
  def transform_OuterMap(self, expr):
    expr.axis = self.transform_if_expr(expr.axis)
    expr.fn = self.transform_expr(expr.fn)
    expr.args = self.transform_expr_tuple(expr.args)
    return expr
  
  def transform_Closure(self, expr):
    expr.args = self.transform_expr_tuple(expr.args)
    expr.fn = self.transform_expr(expr.fn)
    return expr
  
  
  def transform_ClosureElt(self, expr):
    expr.closure = self.transform_expr(expr.closure)
    return expr

  def transform_TypeValue(self, expr):
    pass 
  
  def transform_DelayUntilTyped(self, expr):
    expr.values = self.transform_expr_tuple(expr.values)
    return expr 
  
  def transform_expr(self, expr):
    """Dispatch on the node type and call the appropriate transform method"""

    expr_class = expr.__class__
    if expr_class is Var:
      result = self.transform_Var(expr)
    elif expr_class is Const:
      result = self.transform_Const(expr)
    elif expr_class is Tuple:
      result = self.transform_Tuple(expr)
    elif expr_class is TupleProj:
      result = self.transform_TupleProj(expr)
    elif expr_class is Index:
      result = self.transform_Index(expr)
    elif expr_class is Slice:
      result = self.transform_Slice(expr)
    elif expr_class is Attribute:
      result = self.transform_Attribute(expr)
    elif expr_class is PrimCall:
      result = self.transform_PrimCall(expr)
    elif expr_class is Struct:
      result = self.transform_Struct(expr)
    elif expr_class is AllocArray:
      result = self.transform_AllocArray(expr)
    elif expr_class is Alloc:
      result = self.transform_Alloc(expr)
    elif expr_class is Cast:
      result = self.transform_Cast(expr)
    elif expr_class is ArrayView:
      result = self.transform_ArrayView(expr)
    elif expr_class is TypedFn:
      result = self.transform_TypedFn(expr)
    elif expr_class is UntypedFn:
      result = self.transform_UntypedFn(expr)
    elif expr_class is Call:
      result = self.transform_Call(expr)
    elif expr_class is Map:
      result = self.transform_Map(expr)
    elif expr_class is Reduce:
      result = self.transform_Reduce(expr)
    elif expr_class is Len:
      result = self.transform_Len(expr)
    else:
      method = self.find_method(expr, "transform_")
      if method:
        result = method(expr)

      else:
        assert False, "Unsupported expr %s" % (expr,)
        result = self.transform_generic_expr(expr)
    if result is None:
      return expr 
    else:
      assert isinstance(result, Expr), \
        "Invalid result type in transformation: %s" % (type(result),)
      result.source_info = expr.source_info 
      return result 
  
  def transform_lhs_Var(self, expr):
    return self.transform_Var(expr)

  def transform_lhs_Tuple(self, expr):
    return self.transform_Tuple(expr)

  def transform_lhs_Index(self, expr):
    return self.transform_Index(expr)

  def transform_lhs_Attribute(self, expr):
    return self.transform_Attribute(expr)

  def transform_lhs(self, lhs):
    """
    Overload this is you want different behavior for transformation of left-hand
    side of assignments
    """

    lhs_class = lhs.__class__
    if lhs_class is Var:
      return self.transform_lhs_Var(lhs)
    elif lhs_class is Tuple:
      return self.transform_lhs_Tuple(lhs)
    elif lhs_class is Index:
      return self.transform_lhs_Index(lhs)
    elif lhs_class is Attribute:
      return self.transform_lhs_Attribute(lhs)

    lhs_method = self.find_method(lhs, prefix = "transform_lhs_")
    if lhs_method:
      return lhs_method(lhs)

    method = self.find_method(lhs, prefix = "transform_")
    assert method, "Unknown expression of type %s" % lhs_class
    return method(lhs)

  def transform_expr_list(self, exprs):
    return [self.transform_expr(e) for e in exprs]

  def transform_expr_tuple(self, exprs):
    return tuple(self.transform_expr_list(exprs))

  

  def transform_merge(self, phi_nodes):
    result = {}
    for (k, (left, right)) in phi_nodes.iteritems():
      new_left = self.transform_expr(left)
      new_right = self.transform_expr(right)
      result[k] = new_left, new_right
    return result

  def transform_merge_before_loop(self, phi_nodes):
    return phi_nodes 
  
  def transform_merge_after_loop(self, phi_nodes):
    return self.transform_merge(phi_nodes)
  
  def transform_Assign(self, stmt):
    stmt.rhs = self.transform_expr(stmt.rhs)
    stmt.lhs = self.transform_lhs(stmt.lhs)
    return stmt

  def transform_ExprStmt(self, stmt):
    stmt.value = self.transform_expr(stmt.value)
    return stmt
  
  def transform_Return(self, stmt):
    stmt.value = self.transform_expr(stmt.value)
    return stmt

  def transform_If(self, stmt):
    stmt.true = self.transform_block(stmt.true)
    stmt.false = self.transform_block(stmt.false)
    stmt.merge = self.transform_merge(stmt.merge)
    stmt.cond = self.transform_expr(stmt.cond)
    return stmt

  def transform_While(self, stmt):
    stmt.merge = self.transform_merge_before_loop(stmt.merge)
    stmt.cond = self.transform_expr(stmt.cond)
    stmt.body = self.transform_block(stmt.body)
    stmt.merge = self.transform_merge_after_loop(stmt.merge)
    return stmt
  
  def transform_ForLoop(self, stmt):
    stmt.var = self.transform_expr(stmt.var)
    stmt.merge = self.transform_merge_before_loop(stmt.merge)
    stmt.start = self.transform_expr(stmt.start)
    stmt.stop = self.transform_expr(stmt.stop)
    stmt.step = self.transform_expr(stmt.step)
    stmt.body = self.transform_block(stmt.body)
    stmt.merge = self.transform_merge_after_loop(stmt.merge)
    return stmt 
  
  def transform_Comment(self, stmt):
    return stmt 
  
  def transform_ParFor(self, stmt):
    stmt.fn = self.transform_expr(stmt.fn)
    stmt.bounds= self.transform_expr(stmt.bounds)
    return stmt 
  
  def transform_stmt(self, stmt):
    
    stmt_class = stmt.__class__
    if stmt_class is Assign:
      return self.transform_Assign(stmt)
    elif stmt_class is ForLoop:
      return self.transform_ForLoop(stmt)
    elif stmt_class is While:
      return self.transform_While(stmt)
    elif stmt_class is If:
      return self.transform_If(stmt)
    elif stmt_class is Return:
      return self.transform_Return(stmt)
    elif stmt_class is ExprStmt:
      return self.transform_ExprStmt(stmt)
    elif stmt_class is ParFor:
      return self.transform_ParFor(stmt)
    elif stmt_class is Comment:
      return self.transform_Comment(stmt)
    else:
      assert False, "Unexpected statement %s" % stmt_class

  def transform_block(self, stmts):
    
    self.blocks.push()
    
    if self.reverse: stmts = reversed(stmts)
    
    for old_stmt in stmts:
      new_stmt = self.transform_stmt(old_stmt)
      if new_stmt is not None:
        self.blocks.append_to_current(new_stmt)
    new_block = self.blocks.pop()
    if self.reverse: new_block.reverse()
    return new_block

  def pre_apply(self, old_fn):
    pass  

  def post_apply(self, new_fn):
    pass 

  def apply(self, fn):
    if config.print_transform_timings:
      start_time = time.time()

    transform_name = self.__class__.__name__
    
      
    if config.print_functions_before_transforms == True or \
        (isinstance(config.print_functions_before_transforms, list) and
         transform_name in config.print_functions_before_transforms):
      print
      print "Running transform %s" % transform_name
      print "--- before ---"
      print repr(fn)
      print
    
    self.fn = fn
    self.type_env = fn.type_env

    # push an extra block onto the stack just in case
    # one of the pre_apply methods want to put statements somewhere
    self.blocks.push()
    pre_fn = self.pre_apply(self.fn)
    pre_block = self.blocks.pop()

    if pre_fn is not None:
      fn = pre_fn

    self.fn = fn
    self.type_env = fn.type_env

    # pop the outermost block, which have been written to by
    new_body = self.transform_block(fn.body)
    
    if len(pre_block) > 0:
      new_body = pre_block  + new_body

    fn.body = new_body
    fn.type_env = self.type_env

    self.blocks.push()
    new_fn = self.post_apply(fn)
    post_block = self.blocks.pop()
    if new_fn is None:
      new_fn = fn

    if len(post_block) > 0:
      new_fn.body = new_fn.body + post_block

    if config.print_functions_after_transforms == True or \
        (isinstance(config.print_functions_after_transforms, list) and
         transform_name in config.print_functions_after_transforms):
      print
      print "Done with  %s" % transform_name
      print "--- after ---"
      print repr(new_fn)
      print

    if self.verify:
      try:
        verify(new_fn)
      except:
        print "ERROR after running %s on %s" % (transform_name , new_fn)
        raise

    if config.print_transform_timings:
      end_time = time.time()
      c = self.__class__
      total_time = transform_timings.get(c, 0)
      transform_timings[c] = total_time + (end_time - start_time)
      transform_counts[c] = transform_counts.get(c, 0) + 1
    return new_fn




import numpy as np 
from parakeet.testing_helpers import run_local_tests, expect 

def use_zip(x,y):
  return zip(x,y)

def test_zip_tuples():
  a = (1,2,3)
  b = (np.array([1.0]), np.array([2.0]), np.array([3.0]))
  expect(use_zip, [a,b], tuple(zip(a,b)))

# TODO: 
# Figure out some plan for returning 'dtype' containing
# multiple 
#def test_zip_arrays():
#  a = np.array([1,2,3])
#  b = np.array([True,False,False])
#  expect(use_zip, [a,b], zip(a,b))


if __name__ == '__main__':
  run_local_tests() 




from parakeet import jit, config 
import time 
import numpy as np 
def wald(v, a, rands, u_rands, sigma, accum):
    mu = a / v
    lam = a**2 / sigma**2
    y = rands[:, accum]**2
    x = mu + (mu**2. * y) / (2.*lam) - mu / (2.*lam) * np.sqrt(4.*mu*lam*y + mu**2. * y**2.)
    z = u_rands[:, accum]
    return x


def rep(f, n = 1000, d = 10000):
  rands = np.random.randn(d, 1)
  urands = np.random.rand(d, 1)
  for i in xrange(n):
    f(1,2,rands,urands,1,0)

n = 1000

t = time.time()
rep(wald, n)
py_t = time.time() - t 


waldjit = jit(wald)
#warmup
rep(waldjit, 1)
t = time.time()
rep(waldjit, n)
par_t = time.time() - t 

config.value_specialization = False 
rep(waldjit, 1)
t = time.time()
rep(waldjit, n)
par_t_no_specialization = time.time() - t 

print "Python time per call:", 1000 * (py_t / n), "ms"
print "Parakeet time w/ value specialization:",  1000 * (par_t  / n), "ms"
print "Parakeet time w/out value specialization", 1000 * (par_t_no_specialization  / n), "ms"





from .. import names
from .. analysis import contains_calls, can_inline  
from .. syntax import (If, Assign, While, Return, ForLoop, Var, TypedFn, Const, ExprStmt)

 
from subst import subst_stmt_list
from transform import Transform


def replace_returns(stmts, output_var):
  """Change any returns at the outer scope into assignments to the output var"""

  new_stmts = []
  for stmt in stmts:
    c = stmt.__class__
    if c is Return:
      if output_var:
        new_stmts.append(Assign(output_var, stmt.value))
      continue
    if c is If:
      stmt.true = replace_returns(stmt.true, output_var)
      stmt.false = replace_returns(stmt.false, output_var)
    elif c in (ForLoop, While):
      stmt.body = replace_returns(stmt.body, output_var)

    new_stmts.append(stmt)
  return new_stmts


def replace_return_with_var(body, type_env, return_type):
  result_name = names.fresh("result")
  type_env[result_name] = return_type
  result_var = Var(result_name, type = return_type)
  new_body = replace_returns(body, result_var)
  return result_var, new_body

def do_inline(src_fundef, args, dest_type_env, dest_block, result_var = None):
  arg_names = src_fundef.arg_names
  n_expected = len(arg_names)
  n_given = len(args)
  arg_str = ",".join(src_fundef.arg_names)
  assert n_expected == n_given, \
      "Function %s expects %d args (%s) but given %d (%s)" % \
      (src_fundef.name, n_expected, arg_str, n_given, args)

  rename_dict = {}

  def rename_var(old_name):
    t = src_fundef.type_env[old_name]
    new_name = names.refresh(old_name)
    new_var = Var(new_name, type = t)
    rename_dict[old_name] = new_var
    dest_type_env[new_name] = t

    return new_var

  for (old_formal, actual) in zip(arg_names, args):
    if actual.__class__ in (Var, Const):
      rename_dict[old_formal] = actual
    else:
      new_var = rename_var(old_formal)
      dest_block.append(Assign(new_var, actual))

  for old_name in src_fundef.type_env.iterkeys():
    if old_name not in arg_names:
      rename_var(old_name)

  new_body = subst_stmt_list(src_fundef.body, rename_dict)
  if result_var is None:
    result_var, new_body = \
        replace_return_with_var(new_body, dest_type_env, src_fundef.return_type)
  else:
    new_body = replace_returns(new_body, result_var)
  dest_block.extend(new_body)
  return result_var

class Inliner(Transform):
  def __init__(self):
    Transform.__init__(self)
    self.count = 0


  def transform_TypedFn(self, expr):
    return Inliner().apply(expr)
    #if self.fn.created_by is not None:
    #  return self.fn.created_by.apply(expr)
    #else:
    #  # at the very least, apply high level optimizations
    #  import pipeline
    #  return pipeline.high_level_optimizations.apply(expr)
     
  
  def transform_ExprStmt(self, stmt):

    return Transform.transform_ExprStmt(self, stmt)
  
  def transform_Call(self, expr):

    fn = self.transform_expr(expr.fn)
    target = self.get_fn(fn)
    if target.__class__ is TypedFn:
      closure_args = self.closure_elts(fn)
      if not can_inline(target):
        # print "[Warning] Can't inline %s" % target
        return expr  
      self.count += 1
      curr_block = self.blocks.current()
      combined_args = tuple(closure_args) + tuple(expr.args)
      result_var = do_inline(target, combined_args,
                             self.type_env, curr_block)    
      return result_var
    else:
      return expr


  
  def apply(self, fn):
    if contains_calls(fn):
      return Transform.apply(self, fn)
    else:
      return fn
    

  


import numpy as np 

from ..frontend.decorators import jit  

@jit 
def shuffle(x):
  n = len(x)
  r = np.random.randint(0, n, n)
  for i in xrange(n):
    j = np.fmod(r[i], i)
    old_xj = x[j]
    x[j] = x[i]
    x[i] = old_xj
    
@jit
def permutation(x):
  y = np.zeros_like(x)
  return shuffle(y)

  


import numpy as np
import time 

import parakeet 
from parakeet.testing_helpers import expect_each, run_local_tests


size = (5,5)
float_mat = np.random.uniform(0,1,size=size)
int_mat = np.random.random_integers(0,255,size=size)

matrices = [float_mat, int_mat]

def harris(I):
  m,n = I.shape 
  dx = (I[1:, :] - I[:m-1, :])[:, 1:]
  dy = (I[:, 1:] - I[:, :n-1])[1:, :]
  #
  #   At each point we build a matrix
  #   of derivative products
  #   M =
  #   | A = dx^2     C = dx * dy |
  #   | C = dy * dx  B = dy * dy |
  #
  #   and the score at that point is:
  #      det(M) - k*trace(M)^2
  #
  A = dx * dx
  B = dy * dy
  C = dx * dy
  tr = A + B
  det = A * B - C * C
  k = 0.05
  return det - k * tr * tr


def test_harris():
  expect_each(harris, harris, matrices)


  
if __name__ == '__main__':
  run_local_tests()




import numpy as np 

from parakeet.testing_helpers import run_local_tests, expect 

int_vec = np.array([2,3])
int_mat = np.array([int_vec, int_vec])

bool_vec = np.array([True, True])
bool_mat = np.array([bool_vec, bool_vec])

float32_vec = np.array([1.0, 2.0], dtype='float32')
float32_mat = np.array([float32_vec, float32_vec])

float64_vec = np.array([1.0, 10.0])
float64_mat = np.array([float64_vec, float64_vec])

vecs = [int_vec, bool_vec, float32_vec, float64_vec]
mats = [int_mat, bool_mat, float32_mat, float64_mat]


def unary(fn): 
  for x in mats:
    try:
      expected = fn(x)
    except:
      expected = fn(x.astype('int'))
    if expected.dtype == 'float16':
      expected = fn(x.astype('int'))
    expect(fn, [x], expected, fn.__name__ + "-" + str(x.dtype) + str(len(x.shape)))

def test_negative():
  unary(np.negative)

def test_absolute():
  unary(np.absolute) 
  
def test_rint():
  unary(np.rint)

def test_sign():
  unary(np.sign)
  
def test_conj():
  unary(np.conj)
    
def test_exp():
  unary(np.exp)

def test_exp2():
  unary(np.exp2)
  
def test_log():
  unary(np.log)
  
def test_log2():
  unary(np.log2)

def test_log10():
  unary(np.log10)

def test_expm1():
  unary(np.expm1)
  
def test_log1p():
  unary(np.log1p)
  
def test_sqrt():
  unary(np.sqrt)
  
def test_square():
  unary(np.square)
  
def test_reciprocal():
  unary(np.reciprocal)
  
def test_ones_like():
  unary(np.ones_like)



from ..c_backend.prepare_args import prepare_args 
from ..config import value_specialization 
from ..transforms.pipeline import (flatten, high_level_optimizations, after_indexify, 
                                   final_loop_optimizations) 
from ..value_specialization import specialize


from cuda_compiler import CudaCompiler 

def run(fn, args):
  args = prepare_args(args, fn.input_types)
  fn = after_indexify.apply(fn)
  # fn = flatten(fn)
  fn = final_loop_optimizations.apply(fn)
  if value_specialization:
    fn = specialize(fn, python_values = args)
  compiled_fn = CudaCompiler().compile_entry(fn)
  assert len(args) == len(fn.input_types)
  result = compiled_fn.c_fn(*args)
  return result
  


import numpy as np

from parakeet.testing_helpers import expect, run_local_tests 

def test_zeros_1d():
  expect(np.zeros, [(10,)], np.zeros(10))

def test_zeros_2d():
  expect(np.zeros, [(10,2)], np.zeros((10,2)))  

def test_zeros_1d_int64():
  expect(np.zeros, [(10,), np.int64], np.zeros(10, dtype=np.int64))

def test_zeros_2d_int64():
  expect(np.zeros, [(10,2), np.int64], np.zeros((10,2), dtype=np.int64))

def test_zeros_1d_float64():
  expect(np.zeros, [(10,), np.float64], np.zeros(10, dtype=np.float64))

def test_zeros_2d_float64():
  expect(np.zeros, [(10,2), np.float64], np.zeros((10,2), dtype=np.float64))

def test_zeros_4d_float64():
  expect(np.zeros, [(10,2,2,2), np.float64], np.zeros((10,2,2,2), dtype=np.float64))
  
  
if __name__ == "__main__":
  run_local_tests()
  
  


import numpy as np 

#
# Tensor rotation
# from Peter Mortensen's stackoverflow question 
# @ http://stackoverflow.com/questions/4962606/fast-tensor-rotation-with-numpy/18301915
#

n = 7 
def rotT_loops(T, g):
    Tprime = np.zeros((n,n,n,n))
    for i in range(n):
        for j in range(n):
            for k in range(n):
                for l in range(n):
                    for ii in range(n):
                        for jj in range(n):
                            for kk in range(n):
                                for ll in range(n):
                                    gg = g[ii,i]*g[jj,j]*g[kk,k]*g[ll,l]
                                    Tprime[i,j,k,l] = Tprime[i,j,k,l] + gg*T[ii,jj,kk,ll]
    return Tprime

def rotT_numpy(T, g):
  """
  Accepted response on stack overflow by phillip
  """
  gg = np.outer(g, g)
  gggg = np.outer(gg, gg).reshape(4 * g.shape)
  axes = ((0, 2, 4, 6), (0, 1, 2, 3))
  return np.tensordot(gggg, T, axes)

T = np.random.randn(n,n,n,n)
g = np.random.randn(n,n)

from compare_perf import compare_perf 
compare_perf(rotT_loops, [T, g], extra = {'numpy_tensordot': rotT_numpy})



"""
NMR Pulse propagation 
From http://themodernscientist.com/posts/2013/2013-06-09-simulation_of_nmr_shaped_pulses/
"""

import numpy as np 

pulseLength = 1000. # in microseconds
offset = [-5000., 5000.] # in hertz
n_freq = 500
inputMagnetization = 'Mz' # 'Mx', 'My', or 'Mz'


deltaomega = np.abs(offset[1]-offset[0])/n_freq
relativeomega = np.arange(np.min(offset),np.max(offset),deltaomega)

n_pulse = 1000 # number of points in the pulse, set by user
totalRotation = 180. # in degrees, set by user

fourierCoeffA = np.array([0.49, -1.02, 1.11, -1.57, 0.83, -0.42, 0.26, -0.16, 0.10, -0.07, 0.04, -0.03, 0.01, -0.02, 0.0, -0.01])
x = np.linspace(1,n_pulse,n_pulse)/n_pulse*2.*np.pi
nCosCoef = np.arange(1,len(fourierCoeffA))
cosMat = np.cos(nCosCoef[np.newaxis,:]*x[:,np.newaxis])
cosMat = np.append(np.ones((n_pulse,1)),cosMat,axis=1)*fourierCoeffA
sumMat = np.sum(cosMat,axis=1)

pulseShapeArray = np.zeros((n_pulse,2))
pulseShapeArray[:,0] = np.abs(sumMat)
pulseShapeArray[sumMat<0,1] = 180.

pulseShapeInten = pulseShapeArray[:,0] / np.max(np.abs(pulseShapeArray[:,0]))
pulseShapePhase = pulseShapeArray[:,1] * np.pi/180

xPulseShape = pulseShapeInten * np.cos(pulseShapePhase)
yPulseShape = pulseShapeInten * np.sin(pulseShapePhase)

scalingFactor = np.sum(xPulseShape)/n_pulse
gammaB1max = 1./(pulseLength * 360./totalRotation)/scalingFactor * 1e6
nu1maxdt = 2*np.pi*1e-6*gammaB1max*pulseLength/n_pulse

inputVector = np.array([[0],[0],[1]])
inputMagnetizationDict = {'mx':np.array([[1],[0],[0]]), 'my':np.array([[0],[1],[0]]), 'mz':np.array([[0],[0],[1]]) }
if inputMagnetization.lower() in inputMagnetizationDict.keys():
        inputVector = inputMagnetizationDict[inputMagnetization.lower()]
vectorComponent = inputVector.argmax()

def pulseprop(relativeomega, pulseShapeInten, pulseShapePhase, gammaB1max, nu1maxdt, inputVector, n_pulse, n_freq):
    # Functions for the y and z-rotations and the the function for a generic rotation
    def yrotation(beta):
      return np.array([[np.cos(beta), 0, np.sin(beta)], [0, 1, 0], [-np.sin(beta), 0, np.cos(beta)]])
    def zrotation(beta):
      return np.array([[np.cos(beta), -np.sin(beta), 0], [np.sin(beta), np.cos(beta), 0], [0, 0, 1]])
    def grotation(alpha,theta,phi):
      return np.dot(zrotation(phi),
                    np.dot(yrotation(theta),
                           np.dot(zrotation(alpha), 
                                  np.dot(yrotation(-theta),zrotation(-phi)))))
    
    
    xyzdata = np.zeros((3,len(relativeomega)))
    phi = pulseShapePhase
    
    # Loop through the entire frequency range calculating the rotation matrix (r) at each frequency
    for ind in range(len(relativeomega)):
    
        theta = np.arctan2(pulseShapeInten, relativeomega[ind]/gammaB1max)
        alpha = nu1maxdt * np.sqrt(pulseShapeInten**2+(relativeomega[ind]/gammaB1max)**2)
        
        prop = np.eye(3)
        # The rotation matrix is a recursive loop through each step of the shaped pulse
        for pulseindex in range(n_pulse):
            r = grotation(alpha[pulseindex],theta[pulseindex],phi[pulseindex])
            prop = np.dot(r,prop)
        res = np.dot(prop, inputVector)
        xyzdata[:,ind] = np.dot(prop,inputVector)[:, 0]
        
    return xyzdata

from compare_perf import compare_perf 
from parakeet import jit 
f = jit(pulseprop)
f(relativeomega, pulseShapeInten, pulseShapePhase, gammaB1max, nu1maxdt, inputVector, n_pulse, n_freq)



from ..ndtypes import make_array_type, TupleT, IntT, FnT, ClosureT, increase_rank
from ..syntax import ArrayView, Struct, Expr, ParFor, IndexMap, UntypedFn, TypedFn 
from ..syntax.helpers import zero_i64, get_types, one 

from call_builder import CallBuilder
from adverb_builder import AdverbBuilder

class Builder(AdverbBuilder, CallBuilder):
  

  
  def create_output_array(self, fn, inner_args, 
                          outer_shape = (), 
                          name = "output"):
    if isinstance(outer_shape, (list, tuple)):
      outer_shape = self.tuple(outer_shape)
      
    try:
      inner_shape_tuple = self.call_shape(fn, inner_args)
    except:
      print "Shape inference failed when calling %s with %s" % (fn, inner_args)
      import sys
      print "Error %s ==> %s" % (sys.exc_info()[:2])
      raise

    shape = self.concat_tuples(outer_shape, inner_shape_tuple)
    closure_args = self.closure_elts(fn)
    fn = self.get_fn(fn)
    if isinstance(fn, UntypedFn):
      from .. type_inference import infer_return_type 
      arg_types = get_types(tuple(closure_args) + tuple(inner_args))
      return_type = infer_return_type(fn , arg_types)
      
    else:
      assert isinstance(fn, TypedFn), "Unexpected function %s" % fn
      return_type = self.return_type(fn)
    elt_t = self.elt_type(return_type)
    if len(shape.type.elt_types) > 0:
      return self.alloc_array(elt_t, shape, name)
    else:
      return self.fresh_var(elt_t, name) 
  
  def any_eq(self, tup, target_elt):
    elts = self.tuple_elts(tup)
    is_eq = self.eq(elts[0], target_elt)
    for elt in elts[1:]:
      is_eq = self.or_(is_eq, self.eq(elt, target_elt))
    return is_eq 
  
    
  def parfor(self, fn, bounds):
    assert isinstance(bounds, Expr)
    assert isinstance(bounds.type, (TupleT, IntT))
    assert isinstance(fn, Expr)
    assert isinstance(fn.type, (FnT, ClosureT))
    self.blocks += [ParFor(fn = fn, bounds = bounds)]
  
  def imap(self, fn, bounds):

    assert isinstance(bounds, Expr), "Expected imap bounds to be expression, got %s" % bounds
    if isinstance(bounds.type, TupleT):
      tup = bounds 
      ndims = len(bounds.type.elt_types) 
    else:
      assert isinstance(bounds.type, IntT), \
        "Expected imap bounds to be tuple or int, got %s : %s" % (bounds, bounds.type)
      tup = self.tuple([bounds])
      ndims = 1 
    assert isinstance(fn, Expr), "Expected imap function to be expression, got %s" % (fn,)
    assert isinstance(fn.type, (FnT, ClosureT)), \
      "Expected imap function to have a function type but got %s : %s" % (fn, fn.type)
    elt_type = self.return_type(fn) 
    result_type = increase_rank(elt_type, ndims)
    return IndexMap(fn = fn, shape = tup, type = result_type)
    
    
  def ravel(self, x, explicit_struct = False):
    # TODO: Check the strides to see if any element is equal to 1
    # otherwise do an array_copy
    assert self.is_array(x)
    if x.type.rank == 1:
      return x
    
    nelts = self.nelts(x, explicit_struct = explicit_struct)
    assert isinstance(nelts, Expr)
    shape = self.tuple((nelts,), 'shape', explicit_struct = explicit_struct)
    strides = self.tuple((self.int(1),), "strides", explicit_struct = explicit_struct)
    data = self.attr(x, 'data', 'data_ptr')
    offset = self.attr(x, 'offset')
    t = make_array_type(x.type.elt_type, 1)
    if explicit_struct: 
      return Struct(args = (data, shape, strides, offset, nelts), type = t)
    else:
      return ArrayView(data, shape, strides, offset, nelts, type = t)
  


import numpy as np 
from parakeet import testing_helpers, frontend, analysis, type_inference, transforms

def infer_ranges(python_fn, input_types):
  untyped = frontend.translate_function_value(python_fn)
  typed = type_inference.specialize(untyped, input_types)

  vra = analysis.ValueRangeAnalyis()
  
  vra.visit_fn(typed)
  return vra.ranges 

def expect_const(k, v, c):
  assert v.__class__ is analysis.value_range_analysis.Interval, \
    "Expected %s to be const interval but got %s" % (k,v) 
  assert v.lower == c, "Expected %s to be const value %s not %s" % (k, c,v)
  assert v.upper == c, "Expected %s to be const value %s not %s" % (k, c,v)
  

def expect_range(k, v, lower = None, upper = None):
  assert v.__class__ is analysis.value_range_analysis.Interval, \
    "Expected %s to be interval but got %s" % (k,v)
  if lower is not None:
    assert v.lower == lower, \
      "Expected %s's lower value %s but got %s" % (k, lower,v)
  if upper is not None: 
    assert v.upper == upper, \
      "Expected %s's upper value %s but got %s" % (k, upper,v)

def incr_loop():
  x = 1
  for i in xrange(10):
    x = x + 1
  return x 

def test_incr_loop():
  ranges = infer_ranges(incr_loop, [])
  assert len(ranges) == 4, "Wrong number of variables in %s" % ranges 
  for (k, v) in sorted(ranges.items(), key = lambda (k,v): k):
    if k.startswith("i"):
      expect_range(k, v, 0, 10)
    elif k.startswith("x"):
      assert v.__class__ is analysis.value_range_analysis.Interval, \
        "Expected %s to be interval but got %s" % (k,v)
      assert v.lower in (1,2), "Expected %s to have lower bound of 1 or 2 but got %s" % v 
    else:
      assert False, "Unexpected variables %s = %s" % (k,v)

def const():
  x = 1 
  return x 

def test_const():
  ranges = infer_ranges(const, [])
  assert len(ranges) == 1, "Too many variables: %s" % ranges 
  k,v = ranges.items()[0]
  expect_const(k,v, 1)

def nested_loops():
  x = 0
  y = 0
  z = 0 
  for i in xrange(10):
    x = x + i
    for j in xrange(3,12,1):
      y = y + x
      for k in xrange(3):
        z = z + y
  q = z
  return q

def test_nested_loops():
  ranges = infer_ranges(nested_loops, [])    
  print ranges 
  for k, v in ranges.iteritems():
    if k.startswith("i"):
      expect_range(k,v,0,10)
    elif k.startswith("j"):
      expect_range(k,v,3,12)
    elif k.startswith("x"):
      try:
        expect_range(k,v,0,None)
      except:
        expect_range(k,v,3,np.inf)
    elif k.startswith("q"):
      expect_range(k,v,None,np.inf)
if __name__ == "__main__":
  testing_helpers.run_local_tests()





from ..analysis.collect_vars import collect_binding_names, collect_var_names
from ..analysis.escape_analysis import may_alias

from ..syntax import Return, While, ForLoop, If, Assign, Var, Index, ExprStmt
from transform import Transform

class LoopTransform(Transform):
  def is_simple_block(self, stmts, allow_branches = True):
    for stmt in stmts:
      if stmt.__class__ is If:
        if not allow_branches or \
           not self.is_simple_block(stmt.true) or \
           not self.is_simple_block(stmt.false):
          return False
      elif stmt.__class__ not in (Assign, ExprStmt):
        return False 
    return True

  def pre_apply(self, fn):
    self.may_alias = may_alias(fn)

  def collect_loop_vars(self, loop_vars, loop_body):
    """Gather the variables whose values change between loop iterations"""
    for stmt in loop_body:
      assert stmt.__class__ in (Assign, If, ExprStmt), \
        "Unexpected statement in simple block: %s" % stmt 
      if stmt.__class__ is Assign:
        lhs_names = collect_binding_names(stmt.lhs)
        rhs_names = collect_var_names(stmt.rhs)
        if any(name in loop_vars for name in rhs_names):
          loop_vars.update(lhs_names)

  def collect_memory_accesses(self, loop_body):
      # Assume code is normalized so a read will
      # directly on rhs of an assignment and
      # a write will be directly on the LHS

      # map from variables to index sets
      reads = {}
      writes = {}
      for stmt in loop_body:
        if stmt.__class__ is Assign:
          if stmt.lhs.__class__ is Index:
            assert stmt.lhs.value.__class__ is Var, "Unexpected LHS %s" % stmt.lhs 
            writes.setdefault(stmt.lhs.value.name, set([])).add(stmt.lhs.index)
          if stmt.rhs.__class__ is Index:
            assert stmt.rhs.value.__class__ is Var, "Unexpected LHS %s" % stmt.lhs 
            reads.setdefault(stmt.rhs.value.name, set([])).add(stmt.rhs.index)
      return reads, writes

  def is_loop_var(self, loop_vars, expr):
    #if expr.__class__ is Tuple:
    #  return all(self.is_loop_var(loop_vars, elt) for elt in expr.elts)
    #else:
    return expr.__class__ is Var and expr.name in loop_vars

  def any_loop_vars(self, loop_vars, expr_set):
    return any(self.is_loop_var(loop_vars, expr) for expr in expr_set)
